def test_format_version_matches_library():
    from bioimageio.spec import __version__
    from bioimageio.spec.model import ModelDescr

    lib_format_version = ".".join(__version__.split(".")[:3])
    assert ModelDescr.implemented_format_version == lib_format_version, (
        ModelDescr.implemented_format_version,
        __version__,
    )



# type: ignore
from typing import Optional

import torch
import torch.nn as nn

#
# Model Internal Post-processing
#
# Note: these are mainly for bioimage.io models, where postprocessing has to be done
# inside of the model unless its defined in the general spec


# TODO think about more (multicut-friendly) boundary postprocessing
# e.g. max preserving smoothing: bd = np.maximum(bd, gaussian(bd, sigma=1))
class AccumulateChannels(nn.Module):
    def __init__(self, invariant_channels, accumulate_channels, accumulator):
        super().__init__()
        self.invariant_channels = invariant_channels
        self.accumulate_channels = accumulate_channels
        assert accumulator in ("mean", "min", "max")
        self.accumulator = getattr(torch, accumulator)

    def _accumulate(self, x, c0, c1):
        res = self.accumulator(x[:, c0:c1], dim=1, keepdim=True)
        if not torch.is_tensor(res):
            res = res.values
        assert torch.is_tensor(res)
        return res

    def forward(self, x):
        if self.invariant_channels is None:
            c0, c1 = self.accumulate_channels
            return self._accumulate(x, c0, c1)
        else:
            i0, i1 = self.invariant_channels
            c0, c1 = self.accumulate_channels
            return torch.cat([x[:, i0:i1], self._accumulate(x, c0, c1)], dim=1)


def affinities_to_boundaries(aff_channels, accumulator="max"):
    return AccumulateChannels(None, aff_channels, accumulator)


def affinities_with_foreground_to_boundaries(
    aff_channels, fg_channel=(0, 1), accumulator="max"
):
    return AccumulateChannels(fg_channel, aff_channels, accumulator)


def affinities_to_boundaries2d():
    return affinities_to_boundaries((0, 2))


def affinities_with_foreground_to_boundaries2d():
    return affinities_with_foreground_to_boundaries((1, 3))


def affinities_to_boundaries3d():
    return affinities_to_boundaries((0, 3))


def affinities_with_foreground_to_boundaries3d():
    return affinities_with_foreground_to_boundaries((1, 4))


def affinities_to_boundaries_anisotropic():
    return AccumulateChannels(None, (1, 3), "max")


POSTPROCESSING = {
    "affinities_to_boundaries_anisotropic": affinities_to_boundaries_anisotropic,
    "affinities_to_boundaries2d": affinities_to_boundaries2d,
    "affinities_with_foreground_to_boundaries2d": (
        affinities_with_foreground_to_boundaries2d
    ),
    "affinities_to_boundaries3d": affinities_to_boundaries3d,
    "affinities_with_foreground_to_boundaries3d": (
        affinities_with_foreground_to_boundaries3d
    ),
}


#
# Base Implementations
#


class UNetBase(nn.Module):
    """ """

    def __init__(
        self,
        encoder,
        base,
        decoder,
        out_conv=None,
        final_activation=None,
        postprocessing=None,
    ):
        super().__init__()
        if len(encoder) != len(decoder):
            raise ValueError(
                f"Incompatible depth of encoder (depth={len(encoder)}) and decoder"
                f" (depth={len(decoder)})"
            )

        self.encoder = encoder
        self.base = base
        self.decoder = decoder

        if out_conv is None:
            self.return_decoder_outputs = False
            self._out_channels = self.decoder.out_channels
        elif isinstance(out_conv, nn.ModuleList):
            if len(out_conv) != len(self.decoder):
                raise ValueError(
                    f"Invalid length of out_conv, expected {len(decoder)}, got"
                    f" {len(out_conv)}"
                )
            self.return_decoder_outputs = True
            self._out_channels = [
                None if conv is None else conv.out_channels for conv in out_conv
            ]
        else:
            self.return_decoder_outputs = False
            self._out_channels = out_conv.out_channels
        self.out_conv = out_conv

        self.final_activation = self._get_activation(final_activation)
        self.postprocessing = self._get_postprocessing(postprocessing)

    @property
    def in_channels(self):
        return self.encoder.in_channels

    @property
    def out_channels(self):
        return self._out_channels

    @property
    def depth(self):
        return len(self.encoder)

    def _get_activation(self, activation):
        return_activation = None
        if activation is None:
            return None
        if isinstance(activation, nn.Module):
            return activation
        if isinstance(activation, str):
            return_activation = getattr(nn, activation, None)
        if return_activation is None:
            raise ValueError(f"Invalid activation: {activation}")
        return return_activation()

    def _get_postprocessing(self, postprocessing):
        if postprocessing is None:
            return None
        elif isinstance(postprocessing, nn.Module):
            return postprocessing
        elif postprocessing in POSTPROCESSING:
            return POSTPROCESSING[postprocessing]()
        else:
            raise ValueError(f"Invalid postprocessing: {postprocessing}")

    # load encoder / decoder / base states for pretraining
    def load_encoder_state(self, state):
        self.encoder.load_state_dict(state)

    def load_decoder_state(self, state):
        self.decoder.load_state_dict(state)

    def load_base_state(self, state):
        self.base.load_state_dict(state)

    def _apply_default(self, x):
        self.encoder.return_outputs = True
        self.decoder.return_outputs = False

        x, encoder_out = self.encoder(x)
        x = self.base(x)
        x = self.decoder(x, encoder_inputs=encoder_out[::-1])

        if self.out_conv is not None:
            x = self.out_conv(x)
        if self.final_activation is not None:
            x = self.final_activation(x)
        if self.postprocessing is not None:
            x = self.postprocessing(x)

        return x

    def _apply_with_side_outputs(self, x):
        self.encoder.return_outputs = True
        self.decoder.return_outputs = True

        x, encoder_out = self.encoder(x)
        x = self.base(x)
        x = self.decoder(x, encoder_inputs=encoder_out[::-1])

        x = [x if conv is None else conv(xx) for xx, conv in zip(x, self.out_conv)]
        if self.final_activation is not None:
            x = [self.final_activation(xx) for xx in x]

        if self.postprocessing is not None:
            x = [self.postprocessing(xx) for xx in x]

        # we reverse the list to have the full shape output as first element
        return x[::-1]

    def forward(self, x):
        # cast input data to float, hotfix for modelzoo deployment issues, leaving it here for reference
        # x = x.float()
        if self.return_decoder_outputs:
            return self._apply_with_side_outputs(x)
        else:
            return self._apply_default(x)


def _update_conv_kwargs(kwargs, scale_factor):
    # if the scale factor is a scalar or all entries are the same we don't need to update the kwargs
    if isinstance(scale_factor, int) or scale_factor.count(scale_factor[0]) == len(
        scale_factor
    ):
        return kwargs
    else:  # otherwise set anisotropic kernel
        kernel_size = kwargs.get("kernel_size", 3)
        padding = kwargs.get("padding", 1)

        # bail out if kernel size or padding aren't scalars, because it's
        # unclear what to do in this case
        if not (isinstance(kernel_size, int) and isinstance(padding, int)):
            return kwargs

        kernel_size = tuple(
            1 if factor == 1 else kernel_size for factor in scale_factor
        )
        padding = tuple(0 if factor == 1 else padding for factor in scale_factor)
        kwargs.update({"kernel_size": kernel_size, "padding": padding})
        return kwargs


class Encoder(nn.Module):
    def __init__(
        self,
        features,
        scale_factors,
        conv_block_impl,
        pooler_impl,
        anisotropic_kernel=False,
        **conv_block_kwargs,
    ):
        super().__init__()
        if len(features) != len(scale_factors) + 1:
            raise ValueError(
                "Incompatible number of features {len(features)} and scale_factors"
                " {len(scale_factors)}"
            )

        conv_kwargs = [conv_block_kwargs] * len(scale_factors)
        if anisotropic_kernel:
            conv_kwargs = [
                _update_conv_kwargs(kwargs, scale_factor)
                for kwargs, scale_factor in zip(conv_kwargs, scale_factors)
            ]

        self.blocks = nn.ModuleList(
            [
                conv_block_impl(inc, outc, **kwargs)
                for inc, outc, kwargs in zip(features[:-1], features[1:], conv_kwargs)
            ]
        )
        self.poolers = nn.ModuleList([pooler_impl(factor) for factor in scale_factors])
        self.return_outputs = True

        self.in_channels = features[0]
        self.out_channels = features[-1]

    def __len__(self):
        return len(self.blocks)

    def forward(self, x):
        encoder_out = []
        for block, pooler in zip(self.blocks, self.poolers):
            x = block(x)
            encoder_out.append(x)
            x = pooler(x)

        if self.return_outputs:
            return x, encoder_out
        else:
            return x


class Decoder(nn.Module):
    def __init__(
        self,
        features,
        scale_factors,
        conv_block_impl,
        sampler_impl,
        anisotropic_kernel=False,
        **conv_block_kwargs,
    ):
        super().__init__()
        if len(features) != len(scale_factors) + 1:
            raise ValueError(
                "Incompatible number of features {len(features)} and scale_factors"
                " {len(scale_factors)}"
            )

        conv_kwargs = [conv_block_kwargs] * len(scale_factors)
        if anisotropic_kernel:
            conv_kwargs = [
                _update_conv_kwargs(kwargs, scale_factor)
                for kwargs, scale_factor in zip(conv_kwargs, scale_factors)
            ]

        self.blocks = nn.ModuleList(
            [
                conv_block_impl(inc, outc, **kwargs)
                for inc, outc, kwargs in zip(features[:-1], features[1:], conv_kwargs)
            ]
        )
        self.samplers = nn.ModuleList(
            [
                sampler_impl(factor, inc, outc)
                for factor, inc, outc in zip(scale_factors, features[:-1], features[1:])
            ]
        )
        self.return_outputs = False

        self.in_channels = features[0]
        self.out_channels = features[-1]

    def __len__(self):
        return len(self.blocks)

    # FIXME this prevents traces from being valid for other input sizes, need to find
    # a solution to traceable cropping
    def _crop(self, x, shape):
        shape_diff = [(xsh - sh) // 2 for xsh, sh in zip(x.shape, shape)]
        crop = tuple([slice(sd, xsh - sd) for sd, xsh in zip(shape_diff, x.shape)])
        return x[crop]
        # # Implementation with torch.narrow, does not fix the tracing warnings!
        # for dim, (sh, sd) in enumerate(zip(shape, shape_diff)):
        #     x = torch.narrow(x, dim, sd, sh)
        # return x

    def _concat(self, x1, x2):
        return torch.cat([x1, self._crop(x2, x1.shape)], dim=1)

    def forward(self, x, encoder_inputs):
        if len(encoder_inputs) != len(self.blocks):
            raise ValueError(
                f"Invalid number of encoder_inputs: expect {len(self.blocks)}, got"
                f" {len(encoder_inputs)}"
            )

        decoder_out = []
        for block, sampler, from_encoder in zip(
            self.blocks, self.samplers, encoder_inputs
        ):
            x = sampler(x)
            x = block(self._concat(x, from_encoder))
            decoder_out.append(x)

        if self.return_outputs:
            return decoder_out + [x]
        else:
            return x


def get_norm_layer(norm, dim, channels, n_groups=32):
    if norm is None:
        return None
    if norm == "InstanceNorm":
        return nn.InstanceNorm2d(channels) if dim == 2 else nn.InstanceNorm3d(channels)
    elif norm == "GroupNorm":
        return nn.GroupNorm(min(n_groups, channels), channels)
    elif norm == "BatchNorm":
        return nn.BatchNorm2d(channels) if dim == 2 else nn.BatchNorm3d(channels)
    else:
        raise ValueError(
            "Invalid norm: expect one of 'InstanceNorm', 'BatchNorm' or 'GroupNorm',"
            f" got {norm}"
        )


class ConvBlock(nn.Module):
    def __init__(
        self,
        in_channels,
        out_channels,
        dim,
        kernel_size=3,
        padding=1,
        norm="InstanceNorm",
    ):
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels

        conv = nn.Conv2d if dim == 2 else nn.Conv3d

        if norm is None:
            self.block = nn.Sequential(
                conv(
                    in_channels, out_channels, kernel_size=kernel_size, padding=padding
                ),
                nn.ReLU(inplace=True),
                conv(
                    out_channels, out_channels, kernel_size=kernel_size, padding=padding
                ),
                nn.ReLU(inplace=True),
            )
        else:
            self.block = nn.Sequential(
                get_norm_layer(norm, dim, in_channels),
                conv(
                    in_channels, out_channels, kernel_size=kernel_size, padding=padding
                ),
                nn.ReLU(inplace=True),
                get_norm_layer(norm, dim, out_channels),
                conv(
                    out_channels, out_channels, kernel_size=kernel_size, padding=padding
                ),
                nn.ReLU(inplace=True),
            )

    def forward(self, x):
        return self.block(x)


class Upsampler(nn.Module):
    def __init__(self, scale_factor, in_channels, out_channels, dim, mode):
        super().__init__()
        self.mode = mode
        self.scale_factor = scale_factor

        conv = nn.Conv2d if dim == 2 else nn.Conv3d
        self.conv = conv(in_channels, out_channels, 1)

    def forward(self, x):
        x = nn.functional.interpolate(
            x, scale_factor=self.scale_factor, mode=self.mode, align_corners=False
        )
        x = self.conv(x)
        return x


#
# 2d unet implementations
#


class ConvBlock2d(ConvBlock):
    def __init__(self, in_channels, out_channels, **kwargs):
        super().__init__(in_channels, out_channels, dim=2, **kwargs)


class Upsampler2d(Upsampler):
    def __init__(self, scale_factor, in_channels, out_channels, mode="bilinear"):
        super().__init__(scale_factor, in_channels, out_channels, dim=2, mode=mode)


class UNet2d(UNetBase):
    def __init__(
        self,
        in_channels,
        out_channels,
        depth=4,
        initial_features=32,
        gain=2,
        final_activation=None,
        return_side_outputs=False,
        conv_block_impl=ConvBlock2d,
        pooler_impl=nn.MaxPool2d,
        sampler_impl=Upsampler2d,
        postprocessing=None,
        **conv_block_kwargs,
    ):
        features_encoder = [in_channels] + [
            initial_features * gain**i for i in range(depth)
        ]
        features_decoder = [initial_features * gain**i for i in range(depth + 1)][::-1]
        scale_factors = depth * [2]

        if return_side_outputs:
            if isinstance(out_channels, int) or out_channels is None:
                out_channels = [out_channels] * depth
            if len(out_channels) != depth:
                raise ValueError()
            out_conv: Optional[nn.Module] = nn.ModuleList(
                [
                    nn.Conv2d(feat, outc, 1)
                    for feat, outc in zip(features_decoder[1:], out_channels)
                ]
            )
        else:
            out_conv = (
                None
                if out_channels is None
                else nn.Conv2d(features_decoder[-1], out_channels, 1)
            )

        super().__init__(
            encoder=Encoder(
                features=features_encoder,
                scale_factors=scale_factors,
                conv_block_impl=conv_block_impl,
                pooler_impl=pooler_impl,
                **conv_block_kwargs,
            ),
            decoder=Decoder(
                features=features_decoder,
                scale_factors=scale_factors[::-1],
                conv_block_impl=conv_block_impl,
                sampler_impl=sampler_impl,
                **conv_block_kwargs,
            ),
            base=conv_block_impl(
                features_encoder[-1], features_encoder[-1] * gain, **conv_block_kwargs
            ),
            out_conv=out_conv,
            final_activation=final_activation,
            postprocessing=postprocessing,
        )
        self.init_kwargs = {
            "in_channels": in_channels,
            "out_channels": out_channels,
            "depth": depth,
            "initial_features": initial_features,
            "gain": gain,
            "final_activation": final_activation,
            "return_side_outputs": return_side_outputs,
            "conv_block_impl": conv_block_impl,
            "pooler_impl": pooler_impl,
            "sampler_impl": sampler_impl,
            "postprocessing": postprocessing,
            **conv_block_kwargs,
        }


#
# 3d unet implementations
#


class ConvBlock3d(ConvBlock):
    def __init__(self, in_channels, out_channels, **kwargs):
        super().__init__(in_channels, out_channels, dim=3, **kwargs)


class Upsampler3d(Upsampler):
    def __init__(self, scale_factor, in_channels, out_channels, mode="trilinear"):
        super().__init__(scale_factor, in_channels, out_channels, dim=3, mode=mode)


class AnisotropicUNet(UNetBase):
    def __init__(
        self,
        in_channels,
        out_channels,
        scale_factors,
        initial_features=32,
        gain=2,
        final_activation=None,
        return_side_outputs=False,
        conv_block_impl=ConvBlock3d,
        anisotropic_kernel=False,  # TODO benchmark which option is better and set as default
        postprocessing=None,
        **conv_block_kwargs,
    ):
        depth = len(scale_factors)
        features_encoder = [in_channels] + [
            initial_features * gain**i for i in range(depth)
        ]
        features_decoder = [initial_features * gain**i for i in range(depth + 1)][::-1]

        if return_side_outputs:
            if isinstance(out_channels, int) or out_channels is None:
                out_channels = [out_channels] * depth
            if len(out_channels) != depth:
                raise ValueError()
            out_conv: Optional[nn.Module] = nn.ModuleList(
                [
                    nn.Conv3d(feat, outc, 1)
                    for feat, outc in zip(features_decoder[1:], out_channels)
                ]
            )
        else:
            out_conv = (
                None
                if out_channels is None
                else nn.Conv3d(features_decoder[-1], out_channels, 1)
            )

        super().__init__(
            encoder=Encoder(
                features=features_encoder,
                scale_factors=scale_factors,
                conv_block_impl=conv_block_impl,
                pooler_impl=nn.MaxPool3d,
                anisotropic_kernel=anisotropic_kernel,
                **conv_block_kwargs,
            ),
            decoder=Decoder(
                features=features_decoder,
                scale_factors=scale_factors[::-1],
                conv_block_impl=conv_block_impl,
                sampler_impl=Upsampler3d,
                anisotropic_kernel=anisotropic_kernel,
                **conv_block_kwargs,
            ),
            base=conv_block_impl(
                features_encoder[-1], features_encoder[-1] * gain, **conv_block_kwargs
            ),
            out_conv=out_conv,
            final_activation=final_activation,
            postprocessing=postprocessing,
        )
        self.init_kwargs = {
            "in_channels": in_channels,
            "out_channels": out_channels,
            "scale_factors": scale_factors,
            "initial_features": initial_features,
            "gain": gain,
            "final_activation": final_activation,
            "return_side_outputs": return_side_outputs,
            "conv_block_impl": conv_block_impl,
            "anisotropic_kernel": anisotropic_kernel,
            "postprocessing": postprocessing,
            **conv_block_kwargs,
        }


class UNet3d(AnisotropicUNet):
    def __init__(
        self,
        in_channels,
        out_channels,
        depth=4,
        initial_features=32,
        gain=2,
        final_activation=None,
        return_side_outputs=False,
        conv_block_impl=ConvBlock3d,
        postprocessing=None,
        **conv_block_kwargs,
    ):
        scale_factors = depth * [2]
        super().__init__(
            in_channels,
            out_channels,
            scale_factors,
            initial_features=initial_features,
            gain=gain,
            final_activation=final_activation,
            return_side_outputs=return_side_outputs,
            anisotropic_kernel=False,
            postprocessing=postprocessing,
            conv_block_impl=conv_block_impl,
            **conv_block_kwargs,
        )
        self.init_kwargs = {
            "in_channels": in_channels,
            "out_channels": out_channels,
            "depth": depth,
            "initial_features": initial_features,
            "gain": gain,
            "final_activation": final_activation,
            "return_side_outputs": return_side_outputs,
            "conv_block_impl": conv_block_impl,
            "postprocessing": postprocessing,
            **conv_block_kwargs,
        }



"""script that updates the content of bioimageio/spec/static/spdx_licenses.json with the 'isKnownByZenodo' flag"""

import json
import sys
from pathlib import Path

import requests
from loguru import logger
from tqdm import tqdm

from bioimageio.spec.utils import SpdxLicenses

PROJECT_ROOT = Path(__file__).parent.parent

LICENSES_JSON_FILE = PROJECT_ROOT / "bioimageio/spec/static/spdx_licenses.json"


def main(recheck: bool = False):
    with LICENSES_JSON_FILE.open("rt", encoding="utf-8") as f:
        licenses_full = SpdxLicenses(**json.load(f))

    counts = {True: 0, False: 0, None: 0}
    try:
        for lic in tqdm(licenses_full["licenses"]):
            if recheck is False and isinstance(
                (known := lic.get("isKnownByZenodo")), bool
            ):
                counts[known] += 1
                continue

            url = (
                "https://zenodo.org/api/vocabularies/licenses/"
                + lic["licenseId"].lower()
            )
            r = requests.get(url, timeout=5)

            if 200 <= r.status_code < 300:
                known = True
            elif 400 <= r.status_code < 500:
                known = False
            elif 500 <= r.status_code < 600:
                logger.warning(
                    f"Zenodo server error {r.status_code, r.reason} for {url}"
                )
                known = None
            else:
                raise NotImplementedError(
                    f"got unexpected status code: {r.status_code}"
                )

            # `test_license_zenodo` makes sure `known` is a boolean
            lic["isKnownByZenodo"] = known  # pyright: ignore[reportGeneralTypeIssues]
            counts[known] += 1
    finally:
        with LICENSES_JSON_FILE.open("wt", encoding="utf-8") as f:
            json.dump(licenses_full, f, indent=2)

    print(f"Updated {LICENSES_JSON_FILE}")
    print(f"known to zenodo: {counts}")


if __name__ == "__main__":
    sys.exit(main())



from datetime import datetime
from pathlib import Path
from typing import Any

import pytest
from dateutil.parser import isoparse
from pydantic import PlainSerializer, TypeAdapter
from typing_extensions import Annotated

import bioimageio.spec._internal.io_basics
from bioimageio.spec._internal import types
from bioimageio.spec._internal.io import RelativeFilePath
from bioimageio.spec._internal.types import Datetime, SiUnit
from tests.utils import check_type

TYPE_ARGS = {
    types.Datetime: (2024, 2, 14),
    types.Datetime: datetime.now().isoformat(),
    types.DeprecatedLicenseId: "AGPL-1.0",
    types.Doi: "10.5281/zenodo.5764892",
    types.Identifier: "id",
    types.IdentifierAnno: "id",
    types.ImportantFileSource: "README.md",
    types.LicenseId: "MIT",
    types.LowerCaseIdentifier: "id",
    types.LowerCaseIdentifierAnno: "id",
    types.OrcidId: "0000-0001-2345-6789",
    types.RelativeFilePath: Path(__file__).relative_to(Path().absolute()),
    types.SiUnit: "kg",
    types.AbsoluteDirectory: str(Path(__file__).absolute().parent),
    types.AbsoluteFilePath: str(Path(__file__).absolute()),
    types.FileName: "lala.py",
    types.Version: "1.0",
    types.HttpUrl: "http://example.com",
    bioimageio.spec._internal.io_basics.Sha256: "0" * 64,
}

IGNORE_TYPES_MEMBERS = {
    "AfterValidator",
    "annotated_types",
    "Annotated",
    "annotations",
    "Any",
    "BeforeValidator",
    "ClassVar",
    "datetime",
    "field_validation",
    "FileSource",
    "FormatVersionPlaceholder",  # a literal
    "ImportantFileSource",  # an annoated union
    "iskeyword",
    "isoparse",
    "Literal",
    "NotEmpty",
    "PermissiveFileSource",
    "PlainSerializer",
    "RootModel",
    "S",
    "Sequence",
    "StringConstraints",
    "Type",
    "TypeVar",
    "typing",
    "Union",
    "ValidatedString",
    "YamlValue",
}


@pytest.mark.parametrize(
    "name",
    [
        name
        for name in dir(types)
        if not name.startswith("_") and name not in IGNORE_TYPES_MEMBERS
    ],
)
def test_type_is_usable(name: str):
    """check if a type can be instantiated"""
    typ = getattr(types, name)
    if typ in TYPE_ARGS:
        args = TYPE_ARGS[typ]
        if not isinstance(args, tuple):
            args = (args,)
        _ = typ(*args)
    elif isinstance(typ, str):
        pass  # ignore string constants
    else:
        raise ValueError(f"No idea how to test {name} -> {typ}")


@pytest.mark.parametrize("path", [Path(__file__), Path()])
def test_relative_path(path: Path):
    with pytest.raises(ValueError):
        _ = RelativeFilePath(path.absolute())

    with pytest.raises(ValueError):
        _ = RelativeFilePath(
            str(path.absolute())  # pyright: ignore[reportArgumentType]
        )

    with pytest.raises(ValueError):
        _ = RelativeFilePath(
            path.absolute().as_posix()  # pyright: ignore[reportArgumentType]
        )


@pytest.mark.parametrize("value", ["lx·s", "kg/m^2·s^-2"])
def test_si_unit(value: str):
    check_type(SiUnit, value)


@pytest.mark.parametrize("value", ["lxs", " kg"])
def test_si_unit_invalid(value: str):
    check_type(SiUnit, value, is_invalid=True)


@pytest.mark.parametrize(
    "value,expected",
    [
        (
            "2019-12-11T12:22:32+00:00",
            isoparse("2019-12-11T12:22:32+00:00"),
        ),
        (
            "2019-12-11T12:22:32",
            datetime(2019, 12, 11, 12, 22, 32),
        ),
        (
            "2019-12-11T12:22:32-00:08",
            isoparse("2019-12-11T12:22:32-00:08"),
        ),
    ],
)
def test_datetime(value: str, expected: datetime):
    check_type(
        Datetime,
        value,
        expected_root=expected,
        expected_deserialized=value,
    )


@pytest.mark.parametrize(
    "value",
    [
        "2024-03-06T14:21:34.384830",
        "2024-03-06T14:21:34+00:00",
        "2024-03-06T14:21:34+00:05",
        "2024-03-06T14:21:34-00:08",
        "2019-12-11T12:22:32Z",
    ],
)
def test_datetime_more(value: str):
    from bioimageio.spec._internal.types import (
        _serialize_datetime_json,  # pyright: ignore[reportPrivateUsage]
    )

    root_adapter = TypeAdapter(Datetime)
    datetime_adapter: TypeAdapter[Any] = TypeAdapter(
        Annotated[
            datetime,
            PlainSerializer(_serialize_datetime_json, when_used="json-unless-none"),
        ]
    )

    expected = isoparse(value)

    actual_init = Datetime(expected)
    assert actual_init.root == expected

    actual_root = root_adapter.validate_python(value)
    assert actual_root.root == expected
    assert root_adapter.dump_python(actual_root, mode="python") == expected
    assert root_adapter.dump_python(actual_root, mode="json") == value.replace(
        "Z", "+00:00"
    )

    actual_datetime = datetime_adapter.validate_python(value)
    assert actual_datetime == expected
    assert datetime_adapter.dump_python(actual_datetime, mode="python") == expected
    assert datetime_adapter.dump_python(actual_datetime, mode="json") == value.replace(
        "Z", "+00:00"
    )


@pytest.mark.parametrize(
    "value",
    [
        "2019-12-11T12:22:32+00/00",
        "2019-12-11T12:22:32Y",
        "2019-12-11T12:22:32Zulu",
        "201912-11T12:22:32+00:00",
        "now",
        "today",
    ],
)
def test_datetime_invalid(value: str):
    check_type(Datetime, value, is_invalid=True)



from __future__ import annotations

import collections.abc
import traceback
from abc import ABC, abstractmethod
from copy import deepcopy
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Dict,
    Final,
    FrozenSet,
    Generic,
    List,
    Optional,
    Protocol,
    Tuple,
    Type,
    Union,
    cast,
    get_type_hints,
)

import pydantic
from pydantic import (
    DirectoryPath,
    Field,
    GetCoreSchemaHandler,
    PrivateAttr,
    StringConstraints,
    TypeAdapter,
    model_validator,
)
from pydantic_core import PydanticUndefined, core_schema
from typing_extensions import (
    Annotated,
    LiteralString,
    Self,
    TypeVar,
    TypeVarTuple,
    Unpack,
)

from ..summary import (
    WARNING_LEVEL_TO_NAME,
    ErrorEntry,
    ValidationContextSummary,
    ValidationDetail,
    ValidationSummary,
    WarningEntry,
)
from .field_warning import issue_warning
from .io import BioimageioYamlContent
from .node import Node as Node
from .url import HttpUrl
from .utils import assert_all_params_set_explicitly, get_format_version_tuple
from .validation_context import (
    ValidationContext,
    validation_context_var,
)
from .warning_levels import ALERT, ERROR, INFO


class StringNode(collections.UserString, ABC):
    """deprecated! don't use for new spec fields!"""

    _pattern: ClassVar[str]
    _node_class: Type[Node]
    _node: Optional[Node] = None

    def __init__(self, seq: object) -> None:
        super().__init__(seq)
        type_hints = {
            fn: t
            for fn, t in get_type_hints(self.__class__).items()
            if not fn.startswith("_")
        }
        defaults = {fn: getattr(self.__class__, fn, Field()) for fn in type_hints}
        field_definitions: Dict[str, Any] = {
            fn: (t, defaults[fn]) for fn, t in type_hints.items()
        }
        self._node_class = pydantic.create_model(
            self.__class__.__name__,
            __base__=Node,
            __module__=self.__module__,
            **field_definitions,
        )

        # freeze after initialization
        def __setattr__(self: Self, __name: str, __value: Any):  # type: ignore
            raise AttributeError(f"{self} is immutable.")

        self.__setattr__ = __setattr__  # type: ignore

    @property
    def model_fields(self):
        return self._node_class.model_fields

    def __getattr__(self, name: str):
        if name in self._node_class.model_fields:
            if self._node is None:
                raise AttributeError(f"{name} only available after validation")

            return getattr(self._node, name)

        raise AttributeError(name)

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source: Type[Any], handler: GetCoreSchemaHandler
    ) -> core_schema.CoreSchema:
        assert issubclass(source, StringNode), source
        return core_schema.no_info_after_validator_function(
            cls._validate,
            core_schema.str_schema(pattern=cls._pattern),
            serialization=core_schema.plain_serializer_function_ser_schema(
                cls._serialize,
                info_arg=False,
                return_schema=core_schema.str_schema(),
            ),
        )

    @classmethod
    def _get_data(cls, valid_string_data: str) -> Dict[str, Any]:
        raise NotImplementedError(f"{cls.__name__}._get_data()")

    @classmethod
    def _validate(cls, value: str) -> Self:
        constrained_str_type = Annotated[str, StringConstraints(pattern=cls._pattern)]
        constrained_str_adapter: TypeAdapter[str] = TypeAdapter(constrained_str_type)
        valid_string_data = constrained_str_adapter.validate_python(value)
        data = cls._get_data(valid_string_data)
        self = cls(valid_string_data)
        object.__setattr__(self, "_node", self._node_class.model_validate(data))
        return self

    def _serialize(self) -> str:
        # serialize inner node to call _package when needed
        if self._node is not None:
            _ = self._node.model_dump(mode="json")

        return self.data


SRC = TypeVar("SRC", bound=Union[Node, StringNode])
TGT = TypeVar("TGT", bound=Node)


# converter without any additional args or kwargs:
# class Converter(Generic[SRC, TGT], ABC):
#     # src: ClassVar[Type[SRC]]
#     # tgt: ClassVar[Type[TGT]]
#     # note: the above is not yet possible, see https://github.com/python/typing/discussions/1424
#     # we therefore use an instance
#     def __init__(self, src: Type[SRC], tgt: Type[TGT], /):
#         super().__init__()
#         self.src: Final[Type[SRC]] = src
#         self.tgt: Final[Type[TGT]] = tgt

#     @abstractmethod
#     def _convert(self, src: SRC, tgt: "type[TGT | dict[str, Any]] ", /) -> "TGT | dict[str, Any]":
#         ...

#     def convert(self, source: SRC, /) -> TGT:
#         """convert `source` node

#         Args:
#             source: A bioimageio description node

#         Raises:
#             ValidationError: conversion failed
#         """
#         data = self.convert_as_dict(source)
#         return assert_all_params_set_explicitly(self.tgt)(**data)

#     def convert_as_dict(self, source: SRC) -> Dict[str, Any]:
#         return cast(Dict[str, Any], self._convert(source, dict))


# A TypeVar bound to a TypedDict seemed like a good way to add converter kwargs:
# ```
# class ConverterKwargs(TypedDict):
#     pass
# KW = TypeVar("KW", bound=ConverterKwargs, default=ConverterKwargs)
# ```
# sadly we cannot use a TypeVar bound to TypedDict and then unpack it in the Converter methods,
# see https://github.com/python/typing/issues/1399
# Therefore we use a TypeVarTuple and positional only args instead
# (We are avoiding ParamSpec for its ambiguity 'args vs kwargs')
CArgs = TypeVarTuple("CArgs")


class Converter(Generic[SRC, TGT, Unpack[CArgs]], ABC):
    # src: ClassVar[Type[SRC]]
    # tgt: ClassVar[Type[TGT]]
    # note: the above is not yet possible, see https://github.com/python/typing/discussions/1424
    # we therefore use an instance
    def __init__(self, src: Type[SRC], tgt: Type[TGT], /):
        super().__init__()
        self.src: Final[Type[SRC]] = src
        self.tgt: Final[Type[TGT]] = tgt

    @abstractmethod
    def _convert(
        self, src: SRC, tgt: "type[TGT | dict[str, Any]]", /, *args: Unpack[CArgs]
    ) -> "TGT | dict[str, Any]": ...

    # note: the following is not (yet) allowed, see https://github.com/python/typing/issues/1399
    #       we therefore use `kwargs` (and not `**kwargs`)
    # def convert(self, source: SRC, /, **kwargs: Unpack[KW]) -> TGT:
    def convert(self, source: SRC, /, *args: Unpack[CArgs]) -> TGT:
        """convert `source` node

        Args:
            source: A bioimageio description node

        Raises:
            ValidationError: conversion failed
        """
        data = self.convert_as_dict(source, *args)
        return assert_all_params_set_explicitly(self.tgt)(**data)

    def convert_as_dict(self, source: SRC, /, *args: Unpack[CArgs]) -> Dict[str, Any]:
        return cast(Dict[str, Any], self._convert(source, dict, *args))


class NodeWithExplicitlySetFields(Node):
    fields_to_set_explicitly: ClassVar[FrozenSet[LiteralString]] = frozenset()
    """set set these fields explicitly with their default value if they are not set,
    such that they are always included even when dumping with 'exlude_unset'"""

    @model_validator(mode="before")
    @classmethod
    def set_fields_explicitly(
        cls, data: Union[Any, Dict[str, Any]]
    ) -> Union[Any, Dict[str, Any]]:
        if isinstance(data, dict):
            for name in cls.fields_to_set_explicitly:
                if name not in data:
                    data[name] = cls.model_fields[name].get_default(
                        call_default_factory=True
                    )

        return data


if TYPE_CHECKING:

    class _ResourceDescrBaseAbstractFieldsProtocol(Protocol):
        """workaround to add abstract fields to ResourceDescrBase"""

        # TODO: implement as proper abstract fields of ResourceDescrBase

        type: Any  # should be LiteralString
        format_version: Any  # should be LiteralString

else:

    class _ResourceDescrBaseAbstractFieldsProtocol:
        pass


class ResourceDescrBase(
    NodeWithExplicitlySetFields, ABC, _ResourceDescrBaseAbstractFieldsProtocol
):
    """base class for all resource descriptions"""

    _validation_summary: Optional[ValidationSummary] = None

    fields_to_set_explicitly: ClassVar[FrozenSet[LiteralString]] = frozenset(
        {"type", "format_version"}
    )
    implemented_format_version: ClassVar[str]
    implemented_format_version_tuple: ClassVar[Tuple[int, int, int]]

    # @field_validator("format_version", mode="before", check_fields=False)
    # field_validator on "format_version" is not possible, because we want to use
    #   "format_version" in a descriminated Union higher up
    # (PydanticUserError: Cannot use a mode='before' validator in the discriminator
    #   field 'format_version' of Model 'CollectionDescr')
    @model_validator(mode="before")
    @classmethod
    def _ignore_future_patch(cls, data: Union[Dict[Any, Any], Any], /) -> Any:
        if (
            cls.implemented_format_version == "unknown"
            or not isinstance(data, dict)
            or "format_version" not in data
        ):
            return data

        value = data["format_version"]
        fv = get_format_version_tuple(value)
        if fv is None:
            return data

        if (
            fv[0] == cls.implemented_format_version_tuple[0]
            and fv[1:] > cls.implemented_format_version_tuple[1:]
        ):
            issue_warning(
                "future format_version '{value}' treated as '{implemented}'",
                value=value,
                msg_context=dict(implemented=cls.implemented_format_version),
                severity=ALERT,
            )
            data["format_version"] = cls.implemented_format_version

        return data

    @model_validator(mode="after")
    def _set_init_validation_summary(self) -> Self:
        context = validation_context_var.get()
        self._validation_summary = ValidationSummary(
            name="bioimageio validation",
            source_name=context.source_name,
            type=self.type,
            format_version=self.format_version,
            status="failed" if isinstance(self, InvalidDescr) else "passed",
            details=[
                ValidationDetail(
                    name=f"initialized {self.__class__.__name__} to describe {self.type} {self.implemented_format_version}",
                    status="passed",
                    context=None,  # context for format validation detail is identical
                )
            ],
        )
        return self

    @property
    def validation_summary(self) -> ValidationSummary:
        assert self._validation_summary is not None, "access only after initialization"
        return self._validation_summary

    _root: Union[HttpUrl, DirectoryPath] = PrivateAttr(
        default_factory=lambda: validation_context_var.get().root
    )

    @property
    def root(self) -> Union[HttpUrl, DirectoryPath]:
        return self._root

    @classmethod
    def __pydantic_init_subclass__(cls, **kwargs: Any):
        super().__pydantic_init_subclass__(**kwargs)
        if (
            "format_version" in cls.model_fields
            and cls.model_fields["format_version"].default is not PydanticUndefined
        ):
            cls.implemented_format_version = cls.model_fields["format_version"].default
            if "." not in cls.implemented_format_version:
                cls.implemented_format_version_tuple = (0, 0, 0)
            else:
                fv_tuple = get_format_version_tuple(cls.implemented_format_version)
                assert (
                    fv_tuple is not None
                ), f"failed to cast '{cls.implemented_format_version}' to tuple"
                cls.implemented_format_version_tuple = fv_tuple

    @classmethod
    def load(
        cls, data: BioimageioYamlContent, context: Optional[ValidationContext] = None
    ) -> Union[Self, InvalidDescr]:
        context = context or validation_context_var.get()
        assert isinstance(data, dict)
        with context.replace(log_warnings=False):  # don't log warnings to console
            rd, errors, val_warnings = cls._load_impl(deepcopy(data))

        if context.warning_level > INFO:
            all_warnings_context = context.replace(warning_level=INFO)
            # raise all validation warnings by reloading
            with all_warnings_context:
                _, _, val_warnings = cls._load_impl(deepcopy(data))

        rd.validation_summary.add_detail(
            ValidationDetail(
                errors=errors,
                name=(
                    "bioimageio.spec format validation"
                    f" {rd.type} {cls.implemented_format_version}"
                ),
                status="failed" if errors else "passed",
                warnings=val_warnings,
                context=ValidationContextSummary(
                    perform_io_checks=context.perform_io_checks,
                    known_files=context.known_files,
                    root=str(context.root),
                    warning_level=WARNING_LEVEL_TO_NAME[context.warning_level],
                ),
            )
        )

        return rd

    @classmethod
    def _load_impl(
        cls, data: BioimageioYamlContent
    ) -> Tuple[Union[Self, InvalidDescr], List[ErrorEntry], List[WarningEntry]]:
        rd: Union[Self, InvalidDescr, None] = None
        val_errors: List[ErrorEntry] = []
        val_warnings: List[WarningEntry] = []

        try:
            rd = cls.model_validate(data)
        except pydantic.ValidationError as e:
            for ee in e.errors(include_url=False):
                if (severity := ee.get("ctx", {}).get("severity", ERROR)) < ERROR:
                    val_warnings.append(
                        WarningEntry(
                            loc=ee["loc"],
                            msg=ee["msg"],
                            type=ee["type"],
                            severity=severity,
                        )
                    )
                else:
                    val_errors.append(
                        ErrorEntry(loc=ee["loc"], msg=ee["msg"], type=ee["type"])
                    )

            if len(val_errors) == 0:
                val_errors.append(
                    ErrorEntry(
                        loc=(),
                        msg=(
                            f"Encountered {len(val_warnings)} more severe than warning"
                            " level "
                            f"'{WARNING_LEVEL_TO_NAME[validation_context_var.get().warning_level]}'"
                        ),
                        type="severe_warnings",
                    )
                )
        except Exception as e:
            val_errors.append(
                ErrorEntry(
                    loc=(),
                    msg=str(e),
                    type=type(e).__name__,
                    traceback=traceback.format_tb(e.__traceback__),
                )
            )

        if rd is None:
            try:
                rd = InvalidDescr.model_validate(data)
            except Exception:
                resource_type = cls.model_fields["type"].default
                format_version = cls.implemented_format_version
                rd = InvalidDescr(type=resource_type, format_version=format_version)

        return rd, val_errors, val_warnings


class InvalidDescr(
    ResourceDescrBase,
    extra="allow",
    title="An invalid resource description",
):
    type: Any = "unknown"
    format_version: Any = "unknown"
    fields_to_set_explicitly: ClassVar[FrozenSet[LiteralString]] = frozenset()


class KwargsNode(Node):
    def get(self, item: str, default: Any = None) -> Any:
        return self[item] if item in self else default

    def __getitem__(self, item: str) -> Any:
        if item in self.model_fields:
            return getattr(self, item)
        else:
            raise KeyError(item)

    def __contains__(self, item: str) -> int:
        return item in self.model_fields



from __future__ import annotations

from typing import Any, Dict, Type

import pytest

from bioimageio.spec._internal.common_nodes import Node
from bioimageio.spec._internal.validation_context import ValidationContext
from bioimageio.spec._internal.warning_levels import WARNING
from bioimageio.spec.generic.v0_2 import (
    AttachmentsDescr,
    Author,
    CiteEntry,
    GenericDescr,
    Maintainer,
)
from tests.utils import check_node

EXAMPLE_COM = "https://example.com/"


@pytest.mark.parametrize(
    "node_class,kwargs,is_valid",
    [
        (AttachmentsDescr, dict(magic_number=5), True),
        (AttachmentsDescr, dict(extra=dict(more="of this")), True),
        (AttachmentsDescr, dict(only="other stuff"), True),
        (AttachmentsDescr, dict(files="not a list"), False),
        (AttachmentsDescr, dict(files=["non-existing-file"]), False),
        (AttachmentsDescr, dict(files=[123]), False),
        (Author, dict(name="only_name"), True),
        (
            Author,
            dict(
                name="Me",
                affiliation="Paradise",
                email="them@example.com",
                github_user="ghuser",
                orcid="0000-0001-2345-6789",
            ),
            True,
        ),
        (
            Author,
            dict(
                affiliation="Paradise",
                email="you@example.com",
                github_user="ghuser",
                orcid="0000-0001-2345-6789",
            ),
            False,
        ),
        (
            Author,
            dict(
                name="Me",
                affiliation="Paradise",
                email="me@example.com",
                github_user="ghuser",
                orcid="0000-0001-2345-6788",
            ),
            False,
        ),
        (Author, dict(name=5), False),
        (Maintainer, dict(github_user="ghuser_only"), True),
        (
            Maintainer,
            dict(
                name="Me",
                affiliation="Paradise",
                email="me@example.com",
                github_user="ghuser",
                orcid="0000-0001-2345-6789",
            ),
            True,
        ),
        (Maintainer, dict(name="without_gh"), False),
        (Maintainer, dict(github_user=5.5), False),
        (CiteEntry, dict(text="lala", url=EXAMPLE_COM), True),
        (CiteEntry, dict(text="lala", doi="10.1234fakedoi"), True),
        (
            CiteEntry,
            dict(
                text="Title",
                doi="10.1109/5.771073",
                url="https://ieeexplore.ieee.org/document/771073",
            ),
            True,
        ),
        (CiteEntry, dict(text="lala"), False),
        (CiteEntry, dict(url=EXAMPLE_COM), False),
        (
            GenericDescr,
            dict(
                authors=[{"name": "Me"}],
                description="the description",
                format_version=GenericDescr.implemented_format_version,
                license="BSD-2-Clause-FreeBSD",
                name="my name",
                type="my_type",
                unknown_extra_field="present",
                version="0.1.0",
            ),
            True,
        ),
        (
            GenericDescr,
            dict(
                format_version=GenericDescr.implemented_format_version,
                name="your name",
                description="my description",
                attachments={"something": None},
                type="my_type",
                version="0.1.0",
            ),
            True,
        ),
        (GenericDescr, dict(text="lala"), False),
        (GenericDescr, dict(url=EXAMPLE_COM), False),
        (
            GenericDescr,
            dict(
                authors=[{"name": "Me"}],
                description="the description",
                format_version=GenericDescr.implemented_format_version,
                license="BSD-2-Clause-FreeBSD",
                name="my name",
                type="my_type",
                unknown_extra_field="present",
                version="0.1.0",
            ),
            True,
        ),
        (
            GenericDescr,
            dict(
                format_version=GenericDescr.implemented_format_version,
                name="your name",
                description="my description",
                attachments={"something": 42},
                type="my_type",
                version="0.1.0",
            ),
            True,
        ),
        (
            GenericDescr,
            dict(
                format_version=GenericDescr.implemented_format_version,
                version="0.1.0",
                type="my_type",
                name="their name",
            ),
            False,
        ),
        (
            GenericDescr,
            dict(
                format_version=GenericDescr.implemented_format_version,
                version="0.1.0",
                type="my_type",
                name="its name",
                attachments={"files": ["missing"], "something": 42},
            ),
            False,
        ),
    ],
)
def test_node(node_class: Type[Node], kwargs: Dict[str, Any], is_valid: bool):
    check_node(node_class, kwargs, is_invalid=not is_valid)


def test_deprecated_license_in_generic():
    check_node(
        GenericDescr,
        dict(
            format_version=GenericDescr.implemented_format_version,
            name="my name",
            description="my description",
            authors=[{"name": "Me"}],
            type="my_type",
            version="0.1.0",
            license="BSD-2-Clause-FreeBSD",
            cite=[dict(text="lala", url=EXAMPLE_COM)],
        ),
        context=ValidationContext(warning_level=WARNING),
        is_invalid=True,
    )



from contextlib import nullcontext
from typing import Any, Dict, Tuple

import pytest


@pytest.mark.parametrize(
    "args,kwargs,valid",
    [
        ((1, 1, 1, 1, 1), dict(c1=1, c2=1, d=1), True),
        ((1, 1, 1, 1, 1), dict(c1=1, c2=1), True),
        ((1, 1, 1, 1), dict(c1=1, c2=1, d=1), True),
        ((1, 1, 1, 1), dict(c1=1, c2=1), True),
        ((1, 1, 1, 1), dict(c1=1), False),
        ((1, 1, 1), dict(c1=1, c2=1), False),
        ((1, 1, 1, 1, 1), dict(c2=1), False),
        ((1, 1, 1), dict(b2=1, c1=1, c2=1), True),
        ((1, 1), dict(b1=1, b2=1, c1=1, c2=1), True),
        ((1,), dict(a2=1, b1=1, b2=1, c1=1, c2=1), False),
        ((), dict(a1=1, a2=1, b1=1, b2=1, c1=1, c2=1), False),
    ],
)
def test_assert_all_params_set_explicitly(
    args: Tuple[int, ...], kwargs: Dict[str, int], valid: bool
):
    from bioimageio.spec._internal.utils import assert_all_params_set_explicitly

    def func(
        a1: int = 0,
        a2: int = 0,
        /,
        b1: int = 0,
        b2: int = 0,
        *args: Any,
        c1: int = 0,
        c2: int = 0,
        **kwargs: Any,
    ):
        print(a1, a2, b1, b2, args, c1, c2, kwargs)

    func_explicit = assert_all_params_set_explicitly(func)

    func(*args, **kwargs)

    if valid:
        ctxt = nullcontext()
    else:
        ctxt = pytest.raises(AssertionError)

    with ctxt:
        func_explicit(*args, **kwargs)



from typing import Any, ClassVar, Optional, Type, Union

import pydantic
import requests
import requests.exceptions
from loguru import logger
from pydantic import RootModel, model_validator
from typing_extensions import Literal, assert_never

from .field_warning import issue_warning
from .root_url import RootHttpUrl
from .validation_context import validation_context_var


def _validate_url(url: Union[str, pydantic.HttpUrl]) -> pydantic.AnyUrl:
    return _validate_url_impl(url, request_mode="head")


def _validate_url_impl(
    url: Union[str, pydantic.HttpUrl],
    request_mode: Literal["head", "get_stream", "get"],
    timeout: int = 3,
) -> pydantic.AnyUrl:

    url = str(url)
    val_url = url

    if url.startswith("https://colab.research.google.com/github/"):
        # get requests for colab returns 200 even if the source notebook does not exists.
        # We therefore validate the url to the notebbok instead (for github notebooks)
        val_url = url.replace(
            "https://colab.research.google.com/github/", "https://github.com/"
        )
    elif url.startswith("https://colab.research.google.com/"):
        # TODO: improve validation of non-github colab urls
        issue_warning(
            "colab urls currently pass even if the notebook url was not found. Cannot fully validate {value}",
            value=url,
        )

    try:
        if request_mode == "head":
            response = requests.head(val_url, timeout=timeout)
        elif request_mode == "get_stream":
            response = requests.get(val_url, stream=True, timeout=timeout)
        elif request_mode == "get":
            response = requests.get(val_url, stream=False, timeout=timeout)
        else:
            assert_never(request_mode)
    except (
        requests.exceptions.ChunkedEncodingError,
        requests.exceptions.ContentDecodingError,
        requests.exceptions.InvalidHeader,
        requests.exceptions.InvalidJSONError,
        requests.exceptions.InvalidSchema,
        requests.exceptions.InvalidURL,
        requests.exceptions.MissingSchema,
        requests.exceptions.StreamConsumedError,
        requests.exceptions.TooManyRedirects,
        requests.exceptions.UnrewindableBodyError,
        requests.exceptions.URLRequired,
    ) as e:
        raise ValueError(
            f"Invalid URL '{url}': {e}\nrequest: {e.request}\nresponse: {e.response}"
        )
    except requests.RequestException as e:
        issue_warning(
            "Failed to validate URL '{value}': {error}\nrequest: {request}\nresponse: {response}",
            value=url,
            msg_context={"error": str(e), "response": e.response, "request": e.request},
        )
    except Exception as e:
        issue_warning(
            "Failed to validate URL '{value}': {error}",
            value=url,
            msg_context={"error": str(e)},
        )
    else:
        if response.status_code == 302:  # found
            pass
        elif response.status_code in (301, 303, 308):
            issue_warning(
                "URL redirected ({status_code}): consider updating {value} with new"
                + " location: {location}",
                value=url,
                msg_context={
                    "status_code": response.status_code,
                    "location": response.headers.get("location"),
                },
            )
        elif response.status_code == 403:  # forbidden
            if request_mode == "head":
                return _validate_url_impl(
                    url, request_mode="get_stream", timeout=timeout
                )
            elif request_mode == "get_stream":
                return _validate_url_impl(url, request_mode="get", timeout=timeout)
            elif request_mode == "get":
                raise ValueError(f"{response.status_code}: {response.reason} {url}")
            else:
                assert_never(request_mode)
        elif response.status_code == 405:
            issue_warning(
                "{status_code}: {reason} {value}",
                value=url,
                msg_context={
                    "status_code": response.status_code,
                    "reason": response.reason,
                },
            )
        elif response.status_code != 200:
            raise ValueError(f"{response.status_code}: {response.reason} {url}")

    return pydantic.AnyUrl(url)


class HttpUrl(RootHttpUrl):
    root_model: ClassVar[Type[RootModel[Any]]] = RootModel[pydantic.HttpUrl]
    _exists: Optional[bool] = None

    @model_validator(mode="after")
    def _validate_url(self):
        url = self._validated
        context = validation_context_var.get()
        if context.perform_io_checks and str(url) not in context.known_files:
            self._validated = _validate_url(url)
            self._exists = True

        return self

    def exists(self):
        """True if URL is available"""
        if self._exists is None:
            try:
                self._validated = _validate_url(self._validated)
            except Exception as e:
                logger.info(e)
                self._exists = False
            else:
                self._exists = True

        return self._exists



"""implementation details for building a bioimage.io resource description"""

from typing import Any, Callable, List, Mapping, Optional, Type, TypeVar, Union

from ._internal.common_nodes import InvalidDescr, ResourceDescrBase
from ._internal.io import BioimageioYamlContent
from ._internal.types import FormatVersionPlaceholder
from ._internal.validation_context import ValidationContext, validation_context_var
from .summary import (
    WARNING_LEVEL_TO_NAME,
    ErrorEntry,
    ValidationContextSummary,
    ValidationDetail,
)

ResourceDescrT = TypeVar("ResourceDescrT", bound=ResourceDescrBase)


DISCOVER: FormatVersionPlaceholder = "discover"
"""placeholder for whatever format version an RDF specifies"""


def get_rd_class_impl(
    typ: Any,
    format_version: Any,
    descriptions_map: Mapping[
        Optional[str], Mapping[Optional[str], Type[ResourceDescrT]]
    ],
) -> Type[ResourceDescrT]:
    assert None in descriptions_map
    assert all(None in version_map for version_map in descriptions_map.values())
    assert all(
        fv is None or fv.count(".") == 1
        for version_map in descriptions_map.values()
        for fv in version_map
    )
    if not isinstance(typ, str) or typ not in descriptions_map:
        typ = None

    if (ndots := format_version.count(".")) == 0:
        use_format_version = format_version + ".0"
    elif ndots == 2:
        use_format_version = format_version[: format_version.rfind(".")]
    else:
        use_format_version = None

    descr_versions = descriptions_map[typ]
    return descr_versions.get(use_format_version, descr_versions[None])


def build_description_impl(
    content: BioimageioYamlContent,
    /,
    *,
    context: Optional[ValidationContext] = None,
    format_version: Union[FormatVersionPlaceholder, str] = DISCOVER,
    get_rd_class: Callable[[Any, Any], Type[ResourceDescrT]],
) -> Union[ResourceDescrT, InvalidDescr]:
    context = context or validation_context_var.get()
    errors: List[ErrorEntry] = []
    if isinstance(content, dict):
        for minimum in ("type", "format_version"):
            if minimum not in content:
                errors.append(
                    ErrorEntry(
                        loc=(minimum,), msg=f"Missing field '{minimum}'", type="error"
                    )
                )
            elif not isinstance(content[minimum], str):
                errors.append(
                    ErrorEntry(
                        loc=(minimum,),
                        msg=f"Invalid type '{type(content[minimum])}'",
                        type="error",
                    )
                )
    else:
        errors.append(
            ErrorEntry(
                loc=(), msg=f"Invalid content of type '{type(content)}'", type="error"
            )
        )
        content = {}

    if errors:
        ret = InvalidDescr(**content)  # pyright: ignore[reportArgumentType]
        ret.validation_summary.add_detail(
            ValidationDetail(
                name="extract fields to chose description class",
                status="failed",
                errors=errors,
                context=ValidationContextSummary(
                    perform_io_checks=context.perform_io_checks,
                    known_files=context.known_files,
                    root=str(context.root),
                    warning_level=WARNING_LEVEL_TO_NAME[context.warning_level],
                ),
            )
        )
        return ret

    typ = content["type"]
    rd_class = get_rd_class(typ, content["format_version"])
    rd = rd_class.load(content, context=context)

    if format_version != DISCOVER and not isinstance(rd, InvalidDescr):
        # load with requested format_version
        discover_details = rd.validation_summary.details
        as_rd_class = get_rd_class(typ, format_version)
        rd = as_rd_class.load(content, context=context)
        assert rd.validation_summary is not None
        rd.validation_summary.details[:0] = discover_details

    return rd



from __future__ import annotations

from datetime import datetime
from keyword import iskeyword
from typing import Any, ClassVar, Sequence, Type, TypeVar, Union

import annotated_types
from dateutil.parser import isoparse
from pydantic import PlainSerializer, RootModel, StringConstraints
from typing_extensions import Annotated, Literal

from .constants import DOI_REGEX, SI_UNIT_REGEX
from .io import FileSource as FileSource
from .io import ImportantFileSource as ImportantFileSource
from .io import PermissiveFileSource as PermissiveFileSource
from .io import RelativeFilePath as RelativeFilePath
from .io_basics import AbsoluteDirectory as AbsoluteDirectory
from .io_basics import AbsoluteFilePath as AbsoluteFilePath
from .io_basics import FileName as FileName
from .io_basics import Sha256 as Sha256
from .license_id import DeprecatedLicenseId as DeprecatedLicenseId
from .license_id import LicenseId as LicenseId
from .url import HttpUrl as HttpUrl
from .validated_string import ValidatedString
from .validator_annotations import AfterValidator, BeforeValidator
from .version_type import Version as Version

S = TypeVar("S", bound=Sequence[Any])
NotEmpty = Annotated[S, annotated_types.MinLen(1)]


def _validate_identifier(s: str) -> str:
    if not s.isidentifier():
        raise ValueError(
            f"'{s}' is not a valid (Python) identifier, see"
            + " https://docs.python.org/3/reference/lexical_analysis.html#identifiers"
            + " for details."
        )

    return s


def _validate_is_not_keyword(s: str) -> str:
    if iskeyword(s):
        raise ValueError(f"'{s}' is a Python keyword and not allowed here.")

    return s


def _validate_datetime(dt: Union[datetime, str, Any]) -> datetime:
    if isinstance(dt, datetime):
        return dt
    elif isinstance(dt, str):
        return isoparse(dt)

    raise ValueError(f"'{dt}' not a string or datetime.")


def _validate_orcid_id(orcid_id: str):
    if len(orcid_id) == 19 and all(orcid_id[idx] == "-" for idx in [4, 9, 14]):
        check = 0
        for n in orcid_id[:4] + orcid_id[5:9] + orcid_id[10:14] + orcid_id[15:]:
            # adapted from stdnum.iso7064.mod_11_2.checksum()
            check = (2 * check + int(10 if n == "X" else n)) % 11
        if check == 1:
            return orcid_id  # valid

    raise ValueError(
        f"'{orcid_id} is not a valid ORCID iD in hyphenated groups of 4 digits."
    )


# TODO follow up on https://github.com/pydantic/pydantic/issues/8964
# to remove _serialize_datetime
def _serialize_datetime_json(dt: datetime) -> str:
    return dt.isoformat()


class Datetime(
    RootModel[
        Annotated[
            datetime,
            BeforeValidator(_validate_datetime),
            PlainSerializer(_serialize_datetime_json, when_used="json-unless-none"),
        ]
    ]
):
    """Timestamp in [ISO 8601](#https://en.wikipedia.org/wiki/ISO_8601) format
    with a few restrictions listed [here](https://docs.python.org/3/library/datetime.html#datetime.datetime.fromisoformat).
    """


class Doi(ValidatedString):
    root_model: ClassVar[Type[RootModel[Any]]] = RootModel[
        Annotated[str, StringConstraints(pattern=DOI_REGEX)]
    ]


FormatVersionPlaceholder = Literal["latest", "discover"]
IdentifierAnno = Annotated[
    NotEmpty[str],
    AfterValidator(_validate_identifier),
    AfterValidator(_validate_is_not_keyword),
]


class Identifier(ValidatedString):
    root_model: ClassVar[Type[RootModel[Any]]] = RootModel[IdentifierAnno]


LowerCaseIdentifierAnno = Annotated[IdentifierAnno, annotated_types.LowerCase]


class LowerCaseIdentifier(ValidatedString):
    root_model: ClassVar[Type[RootModel[Any]]] = RootModel[LowerCaseIdentifierAnno]


class OrcidId(ValidatedString):
    """an ORCID identifier"""

    root_model: ClassVar[Type[RootModel[Any]]] = RootModel[
        Annotated[str, AfterValidator(_validate_orcid_id)]
    ]


class SiUnit(ValidatedString):
    """Si unit"""

    root_model: ClassVar[Type[RootModel[Any]]] = RootModel[
        Annotated[
            str,
            StringConstraints(min_length=1, pattern=SI_UNIT_REGEX),
            BeforeValidator(
                lambda s: (
                    s.replace("×", "·").replace("*", "·").replace(" ", "·")
                    if isinstance(s, str)
                    else s
                )
            ),
        ]
    ]



from typing import Union

import pytest

from tests.utils import check_type


@pytest.mark.parametrize(
    "value", [1, "1", 0.1, 1.0, "1.0", "0.0.1", "0.1.0", "0.1.1", "1.0.0", "1.1.1"]
)
def test_version_type(value: Union[str, int, float]):
    from bioimageio.spec._internal.version_type import Version

    check_type(
        Version,
        value,
        expected_root=value,
        expected_deserialized=value,
    )



from datetime import datetime
from typing import Any, Dict, Union

import pytest
from pydantic import ValidationError

from bioimageio.spec._description import validate_format
from bioimageio.spec._internal.root_url import RootHttpUrl
from bioimageio.spec._internal.validation_context import ValidationContext
from bioimageio.spec.model.v0_4 import (
    Author,
    CiteEntry,
    Datetime,
    InputTensorDescr,
    LinkedModel,
    Maintainer,
    ModelDescr,
    OnnxWeightsDescr,
    OutputTensorDescr,
    PostprocessingDescr,
    PreprocessingDescr,
    ScaleLinearKwargs,
    ScaleMeanVarianceDescr,
    ScaleRangeDescr,
    ScaleRangeKwargs,
    TensorName,
    WeightsDescr,
)
from tests.conftest import UNET2D_ROOT
from tests.utils import check_node, check_type, unset


def test_linked_model_lala():
    check_node(
        LinkedModel,
        dict(id="lala"),
        expected_dump_json=dict(id="lala"),
        expected_dump_python=dict(id="lala"),
    )


@pytest.mark.parametrize(
    "kwargs",
    [
        dict(url="https://example.com"),
        dict(id="lala", uri="https://example.com"),
    ],
)
def test_linked_model_invalid(kwargs: Dict[str, Any]):
    check_node(LinkedModel, kwargs, is_invalid=True)


@pytest.mark.parametrize(
    "kwargs,expected",
    [
        (
            dict(source=UNET2D_ROOT / "weights.onnx", sha256="s" * 64),
            dict(source=UNET2D_ROOT / "weights.onnx", sha256="s" * 64),
        ),
        (
            dict(opset_version=5, source=UNET2D_ROOT / "weights.onnx", sha256="s" * 64),
            ValidationError,
        ),
        (
            dict(source=UNET2D_ROOT / "weights.onnx", sha256="s"),
            ValidationError,
        ),
    ],
)
def test_onnx_entry(kwargs: Dict[str, Any], expected: Union[Dict[str, Any], bool]):
    check_node(
        OnnxWeightsDescr,
        kwargs,
        expected_dump_json=expected if isinstance(expected, dict) else unset,
        is_invalid=expected is ValidationError,
        context=ValidationContext(perform_io_checks=False),
    )


VALID_PRE_AND_POSTPROCESSING = [
    dict(name="binarize", kwargs={"threshold": 0.5}),
    dict(name="clip", kwargs={"min": 0.2, "max": 0.5}),
    dict(name="scale_linear", kwargs={"gain": 2.0, "offset": 0.5, "axes": "xy"}),
    dict(name="sigmoid"),
    dict(
        name="zero_mean_unit_variance",
        kwargs={"mode": "fixed", "mean": 1.0, "std": 2.0, "axes": "xy"},
    ),
    dict(name="scale_range", kwargs={"mode": "per_sample", "axes": "xy"}),
    dict(
        name="scale_range",
        kwargs={
            "mode": "per_sample",
            "axes": "xy",
            "min_percentile": 5,
            "max_percentile": 50,
        },
    ),
]

INVALID_PRE_AND_POSTPROCESSING = [
    dict(kwargs={"threshold": 0.5}),
    dict(name="binarize", kwargs={"mode": "fixed", "threshold": 0.5}),
    dict(name="clip", kwargs={"min": "min", "max": 0.5}),
    dict(name="scale_linear", kwargs={"gain": 2.0, "offset": 0.5, "axes": "b"}),
    dict(name="sigmoid", kwargs={"axes": "x"}),
    dict(
        name="zero_mean_unit_variance",
        kwargs={"mode": "unknown", "mean": 1.0, "std": 2.0, "axes": "xy"},
    ),
    dict(name="scale_range", kwargs={"mode": "fixed", "axes": "xy"}),
    dict(
        name="scale_range",
        kwargs={
            "mode": "per_sample",
            "axes": "xy",
            "min_percentile": 50,
            "max_percentile": 50,
        },
    ),
    dict(name="scale_range", kwargs={"mode": "per_sample", "axes": "xy", "min": 0}),
]


@pytest.mark.parametrize(
    "kwargs",
    VALID_PRE_AND_POSTPROCESSING
    + [
        dict(
            name="scale_range",
            kwargs={
                "mode": "per_dataset",
                "axes": "xy",
                "reference_tensor": "some_input_tensor_name",
            },
        ),
    ],
)
def test_preprocessing(kwargs: Dict[str, Any]):
    check_type(PreprocessingDescr, kwargs, expected_deserialized=kwargs)


@pytest.mark.parametrize("kwargs", INVALID_PRE_AND_POSTPROCESSING)
def test_invalid_preprocessing(kwargs: Dict[str, Any]):
    check_type(PreprocessingDescr, kwargs, is_invalid=True)


@pytest.mark.parametrize(
    "kwargs",
    VALID_PRE_AND_POSTPROCESSING
    + [
        dict(name="scale_range", kwargs={"mode": "per_sample", "axes": "xy"}),
        dict(
            name="scale_range",
            kwargs={
                "mode": "per_dataset",
                "axes": "xy",
                "reference_tensor": "some_input_tensor_name",
            },
        ),
        dict(
            name="scale_mean_variance",
            kwargs={"mode": "per_sample", "reference_tensor": "some_tensor_name"},
        ),
        dict(
            name="scale_mean_variance",
            kwargs={"mode": "per_dataset", "reference_tensor": "some_tensor_name"},
        ),
    ],
)
def test_postprocessing(kwargs: Dict[str, Any]):
    check_type(PostprocessingDescr, kwargs, expected_deserialized=kwargs)


@pytest.mark.parametrize(
    "node,expected",
    [
        (
            ScaleRangeDescr(kwargs=ScaleRangeKwargs(mode="per_sample", axes="xy")),
            dict(name="scale_range", kwargs={"mode": "per_sample", "axes": "xy"}),
        ),
        (
            ScaleMeanVarianceDescr(kwargs={"mode": "per_dataset", "reference_tensor": "some_tensor_name"}),  # type: ignore
            dict(
                name="scale_mean_variance",
                kwargs={"mode": "per_dataset", "reference_tensor": "some_tensor_name"},
            ),
        ),
    ],
)
def test_postprocessing_node_input(node: Any, expected: Dict[str, Any]):
    check_type(PostprocessingDescr, node, expected_deserialized=expected)


@pytest.mark.parametrize(
    "kwargs",
    INVALID_PRE_AND_POSTPROCESSING
    + [
        dict(name="scale_mean_variance", kwargs={"mode": "per_sample"}),
        dict(name="scale_mean_variance", kwargs={"mode": "per_dataset"}),
    ],
)
def test_invalid_postprocessing(kwargs: Dict[str, Any]):
    check_type(PostprocessingDescr, kwargs, is_invalid=True)


@pytest.mark.parametrize(
    "kwargs,valid",
    [
        (dict(axes="xy", gain=2.0, offset=0.5), True),
        (dict(offset=2.0), True),
        (dict(gain=2.0), True),
        (dict(axes="xy", gain=[1.0, 2.0], offset=[0.5, 0.3]), True),
        (dict(gain=2.0, offset=0.5), True),
        (dict(), False),  # type: ignore
        (dict(gain=1.0), False),
        (dict(offset=0.0), False),
    ],
)
def test_scale_linear_kwargs(kwargs: Dict[str, Any], valid: bool):
    check_node(ScaleLinearKwargs, kwargs, is_invalid=not valid)


@pytest.mark.parametrize(
    "kwargs",
    [
        {
            "name": "input_1",
            "description": "Input 1",
            "data_type": "float32",
            "axes": "xyc",
            "shape": [128, 128, 3],
            "preprocessing": [
                {
                    "name": "scale_range",
                    "kwargs": {
                        "max_percentile": 99,
                        "min_percentile": 5,
                        "mode": "per_sample",
                        "axes": "xy",
                    },
                }
            ],
        },
        {
            "name": "input_1",
            "description": "Input 1",
            "data_type": "float32",
            "axes": "xyc",
            "shape": [128, 128, 3],
        },
        {
            "name": "tensor_1",
            "data_type": "float32",
            "axes": "xyc",
            "shape": [128, 128, 3],
        },
    ],
)
def test_input_tensor(kwargs: Dict[str, Any]):
    check_node(InputTensorDescr, kwargs)


@pytest.mark.parametrize(
    "kwargs",
    [
        {
            "name": "output_1",
            "description": "Output 1",
            "data_type": "float32",
            "axes": "xyc",
            "shape": [128, 128, 3],
            "postprocessing": [
                {
                    "name": "scale_range",
                    "kwargs": {
                        "max_percentile": 99,
                        "min_percentile": 5,
                        "mode": "per_sample",
                        "axes": "xy",
                    },
                }
            ],
        },
        {
            "name": "output_1",
            "description": "Output 1",
            "data_type": "float32",
            "axes": "xyc",
            "shape": [128, 128, 3],
        },
        {
            "name": "tensor_1",
            "data_type": "float32",
            "axes": "xyc",
            "shape": [128, 128, 3],
        },
    ],
)
def test_output_tensor(kwargs: Dict[str, Any]):
    check_node(OutputTensorDescr, kwargs)


@pytest.fixture
def model_data():
    with ValidationContext(perform_io_checks=False):
        return ModelDescr(
            documentation=UNET2D_ROOT / "README.md",
            license="MIT",
            git_repo="https://github.com/bioimage-io/core-bioimage-io-python",
            description="description",
            authors=[
                Author(name="Author 1", affiliation="Affiliation 1"),
                Author(name="Author 2"),
            ],
            maintainers=[
                Maintainer(
                    name="Maintainer 1",
                    affiliation="Affiliation 1",
                    github_user="fynnbe",
                ),
                Maintainer(github_user="constantinpape"),
            ],
            timestamp=Datetime(datetime.now()),
            cite=[CiteEntry(text="Paper title", url="https://example.com/")],
            inputs=[
                InputTensorDescr(
                    name=TensorName("input_1"),
                    description="Input 1",
                    data_type="float32",
                    axes="xyc",
                    shape=[128, 128, 3],
                ),
            ],
            outputs=[
                OutputTensorDescr(
                    name=TensorName("output_1"),
                    description="Output 1",
                    data_type="float32",
                    axes="xyc",
                    shape=[128, 128, 3],
                ),
            ],
            name="Model",
            tags=[],
            weights=WeightsDescr(
                onnx=OnnxWeightsDescr(source=UNET2D_ROOT / "weights.onnx")
            ),
            test_inputs=[UNET2D_ROOT / "test_input.npy"],
            test_outputs=[UNET2D_ROOT / "test_output.npy"],
            type="model",
        ).model_dump(mode="json")


@pytest.mark.parametrize(
    "update",
    [
        dict(run_mode={"name": "special_run_mode", "kwargs": dict(marathon=True)}),
        dict(name="µ-unicode-model!"),
        dict(weights={"torchscript": {"source": "local_weights"}}),
        dict(weights={"keras_hdf5": {"source": "local_weights"}}),
        dict(weights={"tensorflow_js": {"source": "local_weights"}}),
        dict(weights={"tensorflow_saved_model_bundle": {"source": "local_weights"}}),
        dict(weights={"onnx": {"source": "local_weights"}}),
        dict(
            weights={
                "pytorch_state_dict": {
                    "source": "local_weights",
                    "architecture": "file.py:Model",
                    "architecture_sha256": "0" * 64,
                }
            }
        ),
    ],
)
def test_model(model_data: Dict[str, Any], update: Dict[str, Any]):
    model_data.update(update)
    summary = validate_format(
        model_data, context=ValidationContext(perform_io_checks=False)
    )
    assert summary.status == "passed", summary.format()


def test_warn_long_name(model_data: Dict[str, Any]):
    model_data["name"] = (
        "veeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeery loooooooooooooooong name"
    )
    summary = validate_format(
        model_data, context=ValidationContext(perform_io_checks=False)
    )
    assert summary.status == "passed", summary.format()
    assert summary.details[1].warnings[0].loc == ("name",), summary.format()
    assert summary.details[1].warnings[0].msg == "Name longer than 64 characters."


def test_model_schema_raises_invalid_input_name(model_data: Dict[str, Any]):
    model_data["inputs"][0]["name"] = "invalid/name"
    summary = validate_format(
        model_data,
        context=ValidationContext(
            root=RootHttpUrl("http://example.com/"), perform_io_checks=False
        ),
    )
    assert summary.status == "failed", summary.format()


def test_output_fixed_shape_too_small(model_data: Dict[str, Any]):
    model_data["outputs"] = [
        {
            "name": "output_1",
            "description": "Output 1",
            "data_type": "float32",
            "axes": "xyc",
            "shape": [128, 128, 3],
            "halo": [32, 128, 0],
        }
    ]

    summary = validate_format(
        model_data,
        context=ValidationContext(
            root=RootHttpUrl("http://example.com/"), perform_io_checks=False
        ),
    )
    assert summary.status == "failed", summary.format()


def test_output_ref_shape_mismatch(model_data: Dict[str, Any]):
    model_data["outputs"] = [
        {
            "name": "output_1",
            "description": "Output 1",
            "data_type": "float32",
            "axes": "xyc",
            "shape": {
                "reference_tensor": "input_1",
                "scale": [1, 2, 3, 4],
                "offset": [0, 0, 0, 0],
            },
        }
    ]

    summary = validate_format(
        model_data,
        context=ValidationContext(
            root=RootHttpUrl("http://example.com/"), perform_io_checks=False
        ),
    )
    assert summary.status == "failed", summary.format()


def test_output_ref_shape_too_small(model_data: Dict[str, Any]):
    model_data["outputs"] = [
        {
            "name": "output_1",
            "description": "Output 1",
            "data_type": "float32",
            "axes": "xyc",
            "shape": {
                "reference_tensor": "input_1",
                "scale": [1, 2, 3],
                "offset": [0, 0, 0],
            },
            "halo": [256, 128, 0],
        }
    ]
    summary = validate_format(
        model_data,
        context=ValidationContext(
            root=RootHttpUrl("http://example.com/"), perform_io_checks=False
        ),
    )
    assert summary.status == "failed", summary.format()


def test_model_has_parent_with_id(model_data: Dict[str, Any]):
    model_data["parent"] = dict(id="10.5281/zenodo.5764892")
    summary = validate_format(
        model_data,
        context=ValidationContext(
            root=RootHttpUrl("http://example.com/"), perform_io_checks=False
        ),
    )
    assert summary.status == "passed", summary.format()


def test_model_with_expanded_output(model_data: Dict[str, Any]):
    model_data["outputs"] = [
        {
            "name": "output_1",
            "description": "Output 1",
            "data_type": "float32",
            "axes": "xyzc",
            "shape": dict(
                scale=[1, 1, None, 1],
                offset=[0, 0, 7, 0],
                reference_tensor="input_1",
            ),
        }
    ]

    summary = validate_format(
        model_data,
        context=ValidationContext(
            root=RootHttpUrl("http://example.com/"), perform_io_checks=False
        ),
    )
    assert summary.status == "passed", summary.format()


def test_model_rdf_is_valid_general_rdf(model_data: Dict[str, Any]):
    model_data["type"] = "model_as_generic"
    model_data["format_version"] = "0.2.4"
    summary = validate_format(
        model_data,
        context=ValidationContext(
            root=RootHttpUrl("http://example.com/"), perform_io_checks=False
        ),
    )
    assert summary.status == "passed", summary.format()


def test_model_does_not_accept_unknown_fields(model_data: Dict[str, Any]):
    model_data["unknown_additional_field"] = "shouldn't be here"
    summary = validate_format(
        model_data,
        context=ValidationContext(
            root=RootHttpUrl("http://example.com/"), perform_io_checks=False
        ),
    )
    assert summary.status == "failed", summary.format()



import collections.abc
import re
import shutil
from io import BytesIO
from pathlib import Path
from tempfile import NamedTemporaryFile, mkdtemp
from typing import IO, Dict, Literal, Optional, Sequence, Union, cast
from zipfile import ZIP_DEFLATED

from pydantic import DirectoryPath, FilePath, NewPath

from ._description import InvalidDescr, ResourceDescr, build_description
from ._internal.common_nodes import ResourceDescrBase
from ._internal.io import (
    BioimageioYamlContent,
    BioimageioYamlSource,
    YamlValue,
    download,
    ensure_is_valid_bioimageio_yaml_name,
)
from ._internal.io_basics import BIOIMAGEIO_YAML, AbsoluteFilePath, FileName
from ._internal.io_utils import open_bioimageio_yaml, write_yaml, write_zip
from ._internal.packaging_context import PackagingContext
from ._internal.url import HttpUrl
from ._internal.validation_context import validation_context_var
from ._internal.warning_levels import ERROR
from ._io import load_description
from .model.v0_4 import WeightsFormat


def get_os_friendly_file_name(name: str) -> str:
    return re.sub(r"\W+|^(?=\d)", "_", name)


def get_resource_package_content(
    rd: ResourceDescr,
    /,
    *,
    bioimageio_yaml_file_name: FileName = BIOIMAGEIO_YAML,
    weights_priority_order: Optional[Sequence[WeightsFormat]] = None,  # model only
) -> Dict[FileName, Union[HttpUrl, AbsoluteFilePath, BioimageioYamlContent]]:
    """
    Args:
        rd: resource description
        bioimageio_yaml_file_name: RDF file name
        # for model resources only:
        weights_priority_order: If given, only the first weights format present in the model is included.
                                If none of the prioritized weights formats is found a ValueError is raised.
    """
    os_friendly_name = get_os_friendly_file_name(rd.name)
    bioimageio_yaml_file_name = bioimageio_yaml_file_name.format(
        name=os_friendly_name, type=rd.type
    )

    bioimageio_yaml_file_name = ensure_is_valid_bioimageio_yaml_name(
        bioimageio_yaml_file_name
    )
    content: Dict[FileName, Union[HttpUrl, AbsoluteFilePath]] = {}
    with PackagingContext(
        bioimageio_yaml_file_name=bioimageio_yaml_file_name,
        file_sources=content,
        weights_priority_order=weights_priority_order,
    ):
        rdf_content: BioimageioYamlContent = rd.model_dump(
            mode="json", exclude_unset=True
        )

    _ = rdf_content.pop("rdf_source", None)

    return {**content, bioimageio_yaml_file_name: rdf_content}


def _prepare_resource_package(
    source: Union[BioimageioYamlSource, ResourceDescr],
    /,
    *,
    weights_priority_order: Optional[Sequence[WeightsFormat]] = None,
) -> Dict[FileName, Union[FilePath, BioimageioYamlContent]]:
    """Prepare to package a resource description; downloads all required files.

    Args:
        source: A bioimage.io resource description (as file, raw YAML content or description class)
        context: validation context
        weights_priority_order: If given only the first weights format present in the model is included.
                                If none of the prioritized weights formats is found all are included.
    """
    context = validation_context_var.get()
    bioimageio_yaml_file_name = context.file_name
    if isinstance(source, ResourceDescrBase):
        descr = source
    elif isinstance(source, dict):
        descr = build_description(source)
    else:
        opened = open_bioimageio_yaml(source)
        bioimageio_yaml_file_name = opened.original_file_name
        context = context.replace(
            root=opened.original_root, file_name=opened.original_file_name
        )
        with context:
            descr = build_description(opened.content)

    if isinstance(descr, InvalidDescr):
        raise ValueError(f"{source} is invalid: {descr.validation_summary}")

    with context:
        package_content = get_resource_package_content(
            descr,
            bioimageio_yaml_file_name=bioimageio_yaml_file_name or BIOIMAGEIO_YAML,
            weights_priority_order=weights_priority_order,
        )

    local_package_content: Dict[FileName, Union[FilePath, BioimageioYamlContent]] = {}
    for k, v in package_content.items():
        if not isinstance(v, collections.abc.Mapping):
            v = download(v).path

        local_package_content[k] = v

    return local_package_content


def save_bioimageio_package_as_folder(
    source: Union[BioimageioYamlSource, ResourceDescr],
    /,
    *,
    output_path: Union[NewPath, DirectoryPath, None] = None,
    weights_priority_order: Optional[  # model only
        Sequence[
            Literal[
                "keras_hdf5",
                "onnx",
                "pytorch_state_dict",
                "tensorflow_js",
                "tensorflow_saved_model_bundle",
                "torchscript",
            ]
        ]
    ] = None,
) -> DirectoryPath:
    """Write the content of a bioimage.io resource package to a folder.

    Args:
        source: bioimageio resource description
        output_path: file path to write package to
        weights_priority_order: If given only the first weights format present in the model is included.
                                If none of the prioritized weights formats is found all are included.

    Returns:
        directory path to bioimageio package folder
    """
    package_content = _prepare_resource_package(
        source,
        weights_priority_order=weights_priority_order,
    )
    if output_path is None:
        output_path = Path(mkdtemp())
    else:
        output_path = Path(output_path)

    output_path.mkdir(exist_ok=True, parents=True)
    for name, source in package_content.items():
        if isinstance(source, collections.abc.Mapping):
            write_yaml(cast(YamlValue, source), output_path / name)
        else:
            shutil.copy(source, output_path / name)

    return output_path


def save_bioimageio_package(
    source: Union[BioimageioYamlSource, ResourceDescr],
    /,
    *,
    compression: int = ZIP_DEFLATED,
    compression_level: int = 1,
    output_path: Union[NewPath, FilePath, None] = None,
    weights_priority_order: Optional[  # model only
        Sequence[
            Literal[
                "keras_hdf5",
                "onnx",
                "pytorch_state_dict",
                "tensorflow_js",
                "tensorflow_saved_model_bundle",
                "torchscript",
            ]
        ]
    ] = None,
) -> FilePath:
    """Package a bioimageio resource as a zip file.

    Args:
        rd: bioimageio resource description
        compression: The numeric constant of compression method.
        compression_level: Compression level to use when writing files to the archive.
                           See https://docs.python.org/3/library/zipfile.html#zipfile.ZipFile
        output_path: file path to write package to
        weights_priority_order: If given only the first weights format present in the model is included.
                                If none of the prioritized weights formats is found all are included.

    Returns:
        path to zipped bioimageio package
    """
    package_content = _prepare_resource_package(
        source,
        weights_priority_order=weights_priority_order,
    )
    if output_path is None:
        output_path = Path(
            NamedTemporaryFile(suffix=".bioimageio.zip", delete=False).name
        )
    else:
        output_path = Path(output_path)

    write_zip(
        output_path,
        package_content,
        compression=compression,
        compression_level=compression_level,
    )
    with validation_context_var.get().replace(warning_level=ERROR):
        if isinstance((exported := load_description(output_path)), InvalidDescr):
            raise ValueError(
                f"Exported package '{output_path}' is invalid:"
                + f" {exported.validation_summary}"
            )

    return output_path


def save_bioimageio_package_to_stream(
    source: Union[BioimageioYamlSource, ResourceDescr],
    /,
    *,
    compression: int = ZIP_DEFLATED,
    compression_level: int = 1,
    output_stream: Union[IO[bytes], None] = None,
    weights_priority_order: Optional[  # model only
        Sequence[
            Literal[
                "keras_hdf5",
                "onnx",
                "pytorch_state_dict",
                "tensorflow_js",
                "tensorflow_saved_model_bundle",
                "torchscript",
            ]
        ]
    ] = None,
) -> IO[bytes]:
    """Package a bioimageio resource into a stream.

    Args:
        rd: bioimageio resource description
        compression: The numeric constant of compression method.
        compression_level: Compression level to use when writing files to the archive.
                           See https://docs.python.org/3/library/zipfile.html#zipfile.ZipFile
        output_stream: stream to write package to
        weights_priority_order: If given only the first weights format present in the model is included.
                                If none of the prioritized weights formats is found all are included.

    Note: this function bypasses safety checks and does not load/validate the model after writing.

    Returns:
        stream of zipped bioimageio package
    """
    if output_stream is None:
        output_stream = BytesIO()

    package_content = _prepare_resource_package(
        source,
        weights_priority_order=weights_priority_order,
    )

    write_zip(
        output_stream,
        package_content,
        compression=compression,
        compression_level=compression_level,
    )

    return output_stream



from pathlib import Path

import pytest


def test_load_non_existing_rdf():
    from bioimageio.spec import load_description

    spec_path = Path("some/none/existing/path/to/spec.model.yaml")

    with pytest.raises(FileNotFoundError):
        _ = load_description(spec_path)


@pytest.mark.parametrize(
    "rid",
    [
        "invigorating-lab-coat",
        "invigorating-lab-coat/1",
        "10.5281/zenodo.11092896",  # backup doi of version 1
        "10.5281/zenodo.11092895",  # concept doi of backup
    ],
)
def test_load_by_id(rid: str):
    from bioimageio.spec._internal.io_utils import open_bioimageio_yaml

    rdf = open_bioimageio_yaml(rid).content
    assert rdf["id"] == "invigorating-lab-coat"



from pathlib import Path
from typing import Optional, Union

import pooch
from pydantic import Field
from pydantic_settings import BaseSettings, SettingsConfigDict
from typing_extensions import Annotated


class Settings(BaseSettings, extra="ignore"):
    """environment variables for bioimageio.spec"""

    model_config = SettingsConfigDict(
        env_prefix="BIOIMAGEIO_", env_file=".env", env_file_encoding="utf-8"
    )

    cache_path: Path = pooch.os_cache("bioimageio")
    """bioimageio cache location"""

    id_map: str = (
        "https://uk1s3.embassy.ebi.ac.uk/public-datasets/bioimage.io/id_map.json"
    )
    """url to bioimageio id_map.json to resolve resource IDs.
    """

    id_map_draft: str = (
        "https://uk1s3.embassy.ebi.ac.uk/public-datasets/bioimage.io/id_map_draft.json"
    )
    """url to bioimageio id_map_draft.json to resolve draft IDs ending with '/draft'."""

    resolve_draft: bool = True
    """Flag to resolve draft resource versions following the pattern
    <resource id>/draft.
    Note that anyone may stage a new draft and that such a draft version
    may not have been reviewed yet.
    Set this flag to False to avoid this potential security risk
    and disallow loading draft versions."""

    perform_io_checks: bool = True
    """wether or not to perform validation that requires file io,
    e.g. downloading a remote files.

    Existence of any local absolute file paths is still being checked."""

    log_warnings: bool = True
    """log validation warnings to console"""

    github_username: Optional[str] = None
    """GitHub username for API requests"""

    github_token: Optional[str] = None
    """GitHub token for API requests"""

    CI: Annotated[Union[bool, str], Field(alias="CI")] = False
    """wether or not the execution happens in a continuous integration (CI) environment"""

    user_agent: Optional[str] = None
    """user agent for http requests"""

    @property
    def github_auth(self):
        if self.github_username is None or self.github_token is None:
            return None
        else:
            return (self.github_username, self.github_token)


settings = Settings()



from typing import Union

import pytest
from pydantic import BaseModel, Field, ValidationError


def test_license_id():
    from bioimageio.spec._internal.license_id import LicenseId

    _ = LicenseId("MIT")

    with pytest.raises(ValidationError):
        _ = LicenseId("not_a_valid_license_id")


def test_license_id_in_model():
    """pydantic 2 now defaults to smart unions, which try to find the best match,
    somehow `str` is considered a better match than a valid `LicneseId`, so we should
    only use 'left_to_right' unions with `RootModel[str]`, e.g. `ValidatedString`"""
    from bioimageio.spec._internal.license_id import LicenseId

    class Model(BaseModel):
        lid: Union[LicenseId, str] = Field(union_mode="left_to_right")

    out = Model.model_validate({"lid": "CC-BY-4.0"}).lid
    assert isinstance(out, LicenseId)


def test_deprecated_license_id():
    from bioimageio.spec._internal.license_id import DeprecatedLicenseId

    _ = DeprecatedLicenseId("AGPL-1.0")

    with pytest.raises(ValidationError):
        _ = DeprecatedLicenseId("not_a_valid_license_id")



from typing import Dict, Literal, Optional, TextIO, Union, cast

from loguru import logger
from pydantic import FilePath, NewPath

from bioimageio.spec._internal.io_basics import Sha256

from ._description import (
    DISCOVER,
    InvalidDescr,
    ResourceDescr,
    build_description,
    dump_description,
    ensure_description_is_dataset,
    ensure_description_is_model,
)
from ._internal._settings import settings
from ._internal.common_nodes import ResourceDescrBase
from ._internal.io import BioimageioYamlContent, YamlValue
from ._internal.io_utils import open_bioimageio_yaml, write_yaml
from ._internal.validation_context import validation_context_var
from .common import PermissiveFileSource
from .dataset import AnyDatasetDescr
from .model import AnyModelDescr
from .summary import ValidationSummary


def load_description(
    source: PermissiveFileSource,
    /,
    *,
    format_version: Union[Literal["discover"], Literal["latest"], str] = DISCOVER,
    perform_io_checks: bool = settings.perform_io_checks,
    known_files: Optional[Dict[str, Sha256]] = None,
) -> Union[ResourceDescr, InvalidDescr]:
    """load a bioimage.io resource description

    Args:
        source: Path or URL to an rdf.yaml or a bioimage.io package
                (zip-file with rdf.yaml in it).
        format_version: (optional) Use this argument to load the resource and
                        convert its metadata to a higher format_version.
        perform_io_checks: Wether or not to perform validation that requires file io,
                           e.g. downloading a remote files. The existence of local
                           absolute file paths is still being checked.
        known_files: Allows to bypass download and hashing of referenced files
                     (even if perform_io_checks is True).

    Returns:
        An object holding all metadata of the bioimage.io resource

    """
    if isinstance(source, ResourceDescrBase):
        name = getattr(source, "name", f"{str(source)[:10]}...")
        logger.warning("returning already loaded description '{}' as is", name)
        return source  # pyright: ignore[reportReturnType]

    opened = open_bioimageio_yaml(source)

    context = validation_context_var.get().replace(
        root=opened.original_root,
        file_name=opened.original_file_name,
        perform_io_checks=perform_io_checks,
        known_files=known_files,
    )

    return build_description(
        opened.content,
        context=context,
        format_version=format_version,
    )


def load_model_description(
    source: PermissiveFileSource,
    /,
    *,
    format_version: Union[Literal["discover"], Literal["latest"], str] = DISCOVER,
    perform_io_checks: bool = settings.perform_io_checks,
    known_files: Optional[Dict[str, Sha256]] = None,
) -> AnyModelDescr:
    """same as `load_description`, but addtionally ensures that the loaded
    description is valid and of type 'model'.
    """
    rd = load_description(
        source,
        format_version=format_version,
        perform_io_checks=perform_io_checks,
        known_files=known_files,
    )
    return ensure_description_is_model(rd)


def load_dataset_description(
    source: PermissiveFileSource,
    /,
    *,
    format_version: Union[Literal["discover"], Literal["latest"], str] = DISCOVER,
    perform_io_checks: bool = settings.perform_io_checks,
    known_files: Optional[Dict[str, Sha256]] = None,
) -> AnyDatasetDescr:
    """same as `load_description`, but addtionally ensures that the loaded
    description is valid and of type 'dataset'.
    """
    rd = load_description(
        source,
        format_version=format_version,
        perform_io_checks=perform_io_checks,
        known_files=known_files,
    )
    return ensure_description_is_dataset(rd)


def save_bioimageio_yaml_only(
    rd: Union[ResourceDescr, BioimageioYamlContent, InvalidDescr],
    /,
    file: Union[NewPath, FilePath, TextIO],
):
    """write the metadata of a resource description (`rd`) to `file`
    without writing any of the referenced files in it.

    Note: To save a resource description with its associated files as a package,
    use `save_bioimageio_package` or `save_bioimageio_package_as_folder`.
    """
    if isinstance(rd, ResourceDescrBase):
        content = dump_description(rd)
    else:
        content = rd

    write_yaml(cast(YamlValue, content), file)


def load_description_and_validate_format_only(
    source: PermissiveFileSource,
    /,
    *,
    format_version: Union[Literal["discover"], Literal["latest"], str] = DISCOVER,
    perform_io_checks: bool = settings.perform_io_checks,
    known_files: Optional[Dict[str, Sha256]] = None,
) -> ValidationSummary:
    """load a bioimage.io resource description

    Args:
        source: Path or URL to an rdf.yaml or a bioimage.io package
                (zip-file with rdf.yaml in it).
        format_version: (optional) Use this argument to load the resource and
                        convert its metadata to a higher format_version.
        perform_io_checks: Wether or not to perform validation that requires file io,
                           e.g. downloading a remote files. The existence of local
                           absolute file paths is still being checked.
        known_files: Allows to bypass download and hashing of referenced files
                     (even if perform_io_checks is True).

    Returns:
        Validation summary of the bioimage.io resource found at `source`.

    """
    rd = load_description(
        source,
        format_version=format_version,
        perform_io_checks=perform_io_checks,
        known_files=known_files,
    )
    assert rd.validation_summary is not None
    return rd.validation_summary



from __future__ import annotations

from typing import (
    Any,
    Dict,
    Optional,
    Union,
)

import pydantic
from typing_extensions import (
    Self,
)

from .validation_context import ValidationContext, validation_context_var


class Node(
    pydantic.BaseModel,
    extra="forbid",
    frozen=False,
    populate_by_name=True,
    revalidate_instances="never",
    validate_assignment=True,
    validate_default=False,
    validate_return=True,  # TODO: check if False here would bring a speedup and can still be safe
    use_attribute_docstrings=True,
):
    """Subpart of a resource description"""

    @classmethod
    def model_validate(
        cls,
        obj: Union[Any, Dict[str, Any]],
        *,
        strict: Optional[bool] = None,
        from_attributes: Optional[bool] = None,
        context: Union[ValidationContext, Dict[str, Any], None] = None,
    ) -> Self:
        """Validate a pydantic model instance.

        Args:
            obj: The object to validate.
            strict: Whether to raise an exception on invalid fields.
            from_attributes: Whether to extract data from object attributes.
            context: Additional context to pass to the validator.

        Raises:
            ValidationError: If the object failed validation.

        Returns:
            The validated description instance.
        """
        __tracebackhide__ = True

        if context is None:
            context = validation_context_var.get()
        elif isinstance(context, dict):
            context = ValidationContext(**context)

        if isinstance(obj, dict):
            assert all(isinstance(k, str) for k in obj), obj

        with context:
            # use validation context as context manager for equal behavior of __init__ and model_validate
            return super().model_validate(
                obj, strict=strict, from_attributes=from_attributes
            )



# autogen: start
"""
implementaions of all released minor versions are available in submodules:
- generic v0_2: `bioimageio.spec.generic.v0_2.GenericDescr`
- generic v0_3: `bioimageio.spec.generic.v0_3.GenericDescr`
"""
from typing import Union

from pydantic import Discriminator
from typing_extensions import Annotated

from .v0_2 import GenericDescr as GenericDescr_v0_2
from .v0_3 import GenericDescr as GenericDescr
from .v0_3 import GenericDescr as GenericDescr_v0_3

AnyGenericDescr = Annotated[
    Union[GenericDescr_v0_2, GenericDescr_v0_3], Discriminator("format_version")
]
"""Union of any released generic desription"""
# autogen: stop



from typing import Any, ClassVar, Type

from pydantic import GetCoreSchemaHandler, RootModel
from pydantic_core.core_schema import (
    CoreSchema,
    no_info_after_validator_function,
)


class ValidatedString(str):
    root_model: ClassVar[Type[RootModel[Any]]] = RootModel[str]
    """the pydantic root model to validate the string"""
    # TODO: should we use a TypeAdapter instead?
    # TODO: with future py version:  RootModel[Any] -> RootModel[str | "literal string type"]
    _validated: Any  # pyright: ignore[reportUninitializedInstanceVariable]  # initalized in __new__

    def __new__(cls, object: object):
        self = super().__new__(cls, object)
        self._validated = cls.root_model.model_validate(str(self)).root
        return self

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> CoreSchema:
        return no_info_after_validator_function(cls, handler(str))



from typing import Any


def test_converter_with_arg():
    from bioimageio.spec._internal.common_nodes import Converter, Node

    class A(Node):
        a: int

    class B(Node):
        b: str

    class AtoB(Converter[A, B, str]):
        def _convert(
            self, src: A, tgt: "type[B] | type[dict[str, Any]]", /, prefix: str
        ) -> "B | dict[str, Any]":
            return tgt(b=prefix + str(src.a))

    converter = AtoB(A, B)

    _ = converter.convert(A(a=5), "prefix")



import json
from typing import List, TypedDict

from ._description import ensure_description_is_dataset as ensure_description_is_dataset
from ._description import ensure_description_is_model as ensure_description_is_model
from ._internal.io import download as download
from ._internal.io import extract_file_name as extract_file_name
from ._internal.io import get_sha256 as get_sha256
from ._internal.io import (
    identify_bioimageio_yaml_file_name as identify_bioimageio_yaml_file_name,
)
from ._internal.io import is_valid_bioimageio_yaml_name as is_valid_bioimageio_yaml_name
from ._internal.io_utils import load_array as load_array
from ._internal.io_utils import save_array as save_array
from ._internal.utils import files


class SpdxLicenseEntry(TypedDict):
    isDeprecatedLicenseId: bool
    isKnownByZenodo: bool
    isOsiApproved: bool
    licenseId: str
    name: str
    reference: str


class SpdxLicenses(TypedDict):
    licenseListVersion: str
    licenses: List[SpdxLicenseEntry]
    releaseDate: str


def get_spdx_licenses() -> SpdxLicenses:
    """get details of the SPDX licenses known to bioimageio.spec"""
    with files("bioimageio.spec").joinpath("static/spdx_licenses.json").open(
        "r", encoding="utf-8"
    ) as f:
        return json.load(f)



from pprint import pprint
from typing import Dict, Tuple, Union

import numpy as np

dtype_limits: Dict[str, Tuple[Union[float, int], Union[float, int]]] = {}

for dtype in ["float32", "float64"]:
    info = np.finfo(dtype)
    dtype_limits[dtype] = (float(info.min), float(info.max))


for dtype in ["uint8", "int8", "uint16", "int16", "uint32", "int32", "uint64", "int64"]:
    info = np.iinfo(dtype)
    dtype_limits[dtype] = (info.min, info.max)

if __name__ == "__main__":
    pprint(dtype_limits)



from datetime import datetime
from typing import Any, Dict, Union

import pytest
from pydantic import RootModel, ValidationError

from bioimageio.spec import validate_format
from bioimageio.spec._internal.io import FileDescr
from bioimageio.spec._internal.license_id import LicenseId
from bioimageio.spec._internal.url import HttpUrl
from bioimageio.spec._internal.validation_context import ValidationContext
from bioimageio.spec.model.v0_5 import (
    Author,
    AxisId,
    BatchAxis,
    ChannelAxis,
    CiteEntry,
    Datetime,
    Identifier,
    InputAxis,
    InputTensorDescr,
    IntervalOrRatioDataDescr,
    Maintainer,
    ModelDescr,
    OnnxWeightsDescr,
    OutputTensorDescr,
    SpaceInputAxis,
    SpaceOutputAxis,
    TensorDescrBase,
    TensorId,
    WeightsDescr,
)
from tests.conftest import UNET2D_ROOT
from tests.utils import check_node, check_type


@pytest.mark.parametrize(
    "kwargs",
    [
        dict(
            id="t0",
            test_tensor={"source": UNET2D_ROOT / "test_input.npy"},
            data={"values": ["cat", "dog", "parrot"]},
            axes=[{"type": "channel", "channel_names": ["animal"]}],
        ),
        dict(
            id="t1",
            test_tensor={"source": UNET2D_ROOT / "test_input.npy"},
            data=[
                {"values": ["cat", "dog", "parrot"]},
                {"values": ["mouse", "zebra", "elephant"]},
            ],
            axes=[{"type": "channel", "channel_names": ["animal", "other_animal"]}],
        ),
        dict(
            id="t2",
            test_tensor={"source": UNET2D_ROOT / "test_input.npy"},
            data=[
                {"values": [1, 2, 3]},
                {"type": "uint8"},
            ],
            axes=[
                {"type": "channel", "channel_names": ["animal_code", "animal_count"]}
            ],
        ),
        pytest.param(
            dict(
                id="t3",
                test_tensor={"source": UNET2D_ROOT / "test_input.npy"},
                data=[
                    {"values": ["mouse", "zebra", "elephant"]},
                    {"type": "uint8"},
                ],
                axes=[{"type": "channel", "channel_names": ["animal_code", "count"]}],
            ),
            id="string values and uint data type",
        ),
    ],
)
def test_tensor_base(kwargs: Dict[str, Any]):
    check_node(
        TensorDescrBase, kwargs, context=ValidationContext(perform_io_checks=False)
    )


@pytest.mark.parametrize(
    "kwargs",
    [
        pytest.param(
            dict(
                id="t5",
                test_tensor={"source": UNET2D_ROOT / "test_input.npy"},
                data=[
                    {"values": ["cat", "dog", "parrot"]},
                    {"values": [1.1, 2.2, 3.3]},
                ],
            ),
            id="str and float values",
        ),
        pytest.param(
            dict(
                id="t7",
                test_tensor={"source": UNET2D_ROOT / "test.npy"},
                data=[
                    {"values": ["mouse", "zebra", "elephant"]},
                    {"type": "int8"},
                ],
            ),
            id="string values and int data type",
        ),
    ],
)
def test_tensor_base_invalid(kwargs: Dict[str, Any]):
    check_node(
        TensorDescrBase,
        kwargs,
        is_invalid=True,
        context=ValidationContext(perform_io_checks=False),
    )


@pytest.mark.parametrize(
    "kwargs",
    [
        {
            "id": "input_1",
            "description": "Input 1",
            "data": {"type": "float32"},
            "axes": [
                dict(type="space", id="x", size=10),
                dict(type="space", id="y", size=11),
                dict(type="channel", channel_names=tuple("abc")),
            ],
            "preprocessing": [
                {
                    "id": "scale_range",
                    "kwargs": {
                        "max_percentile": 99,
                        "min_percentile": 5,
                        "axes": ("x", "y"),
                    },
                }
            ],
            "test_tensor": {"source": UNET2D_ROOT / "test_input.npy"},
        },
    ],
)
def test_input_tensor(kwargs: Dict[str, Any]):
    check_node(
        InputTensorDescr, kwargs, context=ValidationContext(perform_io_checks=False)
    )


@pytest.mark.parametrize(
    "kwargs",
    [
        pytest.param(
            dict(
                id="input_2",
                test_tensor={"source": UNET2D_ROOT / "test.npy"},
                data=[
                    {"values": ["cat", "dog", "parrot"]},
                    {"values": ["mouse", "zebra", "elephant"]},
                ],
                axes=[{"type": "channel", "channel_names": ["a", "b", "c"]}],
            ),
            id="channel mismatch",
        ),
    ],
)
def test_input_tensor_invalid(kwargs: Dict[str, Any]):
    check_node(
        InputTensorDescr,
        kwargs,
        is_invalid=True,
        context=ValidationContext(perform_io_checks=False),
    )


@pytest.mark.skip(
    "possibly bug in pydantic? in some envs it passes, in ohters not"
)  # TODO: fix
def test_input_tensor_error_count(model_data: Dict[str, Any]):
    """this test checks that the discrminated union for `InputAxis` does its
    thing and we don't get errors for all options"""
    from bioimageio.spec import build_description

    model_data["inputs"][0]["axes"][0]["id"] = 42
    invalid = build_description(
        model_data, context=ValidationContext(perform_io_checks=False)
    )
    assert (
        len(invalid.validation_summary.details[1].errors) == 1
    ), invalid.validation_summary.details[1].errors


@pytest.mark.parametrize(
    "kwargs",
    [{}, {"type": "batch"}],
)
def test_batch_axis(kwargs: Dict[str, Any]):
    check_node(
        BatchAxis,
        kwargs,
        expected_dump_python={
            "type": "batch",
            "name": "batch",
            "description": "",
            "size": None,
        },
    )


@pytest.mark.parametrize(
    "kwargs",
    [
        {"type": "space", "id": "x", "size": 10},
        SpaceInputAxis(id=AxisId("x"), size=10),
        {"type": "batch"},
    ],
)
def test_input_axis(kwargs: Union[Dict[str, Any], SpaceInputAxis]):
    check_type(InputAxis, kwargs)


@pytest.fixture
def model_data():
    with ValidationContext(perform_io_checks=False):
        model = ModelDescr(
            documentation=UNET2D_ROOT / "README.md",
            license=LicenseId("MIT"),
            git_repo=HttpUrl("https://github.com/bioimage-io/core-bioimage-io-python"),
            format_version="0.5.3",
            description="description",
            authors=[
                Author(name="Author 1", affiliation="Affiliation 1"),
                Author(name="Author 2"),
            ],
            maintainers=[
                Maintainer(
                    name="Maintainer 1",
                    affiliation="Affiliation 1",
                    github_user="fynnbe",
                ),
                Maintainer(github_user="githubuser2"),
            ],
            timestamp=Datetime(datetime.now()),
            cite=[CiteEntry(text="Paper title", url=HttpUrl("https://example.com/"))],
            inputs=[
                InputTensorDescr(
                    id=TensorId("input_1"),
                    description="Input 1",
                    data=IntervalOrRatioDataDescr(type="float32"),
                    axes=[
                        BatchAxis(),
                        ChannelAxis(channel_names=[Identifier("intensity")]),
                        SpaceInputAxis(id=AxisId("x"), size=512),
                        SpaceInputAxis(id=AxisId("y"), size=512),
                    ],
                    test_tensor=FileDescr(source=UNET2D_ROOT / "test_input.npy"),
                ),
            ],
            outputs=[
                OutputTensorDescr(
                    id=TensorId("output_1"),
                    description="Output 1",
                    axes=[
                        BatchAxis(),
                        ChannelAxis(channel_names=[Identifier("intensity")]),
                        SpaceOutputAxis(id=AxisId("x"), size=512),
                        SpaceOutputAxis(id=AxisId("y"), size=512),
                    ],
                    test_tensor=FileDescr(source=UNET2D_ROOT / "test_output.npy"),
                ),
            ],
            name="Model",
            tags=[],
            weights=WeightsDescr(
                onnx=OnnxWeightsDescr(
                    source=UNET2D_ROOT / "weights.onnx", opset_version=15
                )
            ),
            type="model",
        )
        data = model.model_dump(mode="json")
        assert data["documentation"] == str(UNET2D_ROOT / "README.md"), (
            data["documentation"],
            str(UNET2D_ROOT / "README.md"),
        )
        return data


@pytest.mark.parametrize(
    "update",
    [
        dict(run_mode={"name": "special_run_mode", "kwargs": dict(marathon=True)}),
        dict(
            weights={
                "torchscript": {
                    "source": UNET2D_ROOT / "weights.onnx",
                    "pytorch_version": 1.15,
                }
            }
        ),
        dict(
            weights={
                "keras_hdf5": {
                    "source": UNET2D_ROOT / "weights.onnx",
                    "tensorflow_version": 1.10,
                }
            }
        ),
        dict(
            weights={
                "tensorflow_js": {
                    "source": UNET2D_ROOT / "weights.onnx",
                    "tensorflow_version": 1.10,
                }
            }
        ),
        dict(
            weights={
                "tensorflow_saved_model_bundle": {
                    "source": UNET2D_ROOT / "weights.onnx",
                    "tensorflow_version": 1.10,
                }
            }
        ),
        dict(
            weights={
                "onnx": {"source": UNET2D_ROOT / "weights.onnx", "opset_version": 15}
            }
        ),
        dict(
            weights={
                "pytorch_state_dict": {
                    "source": UNET2D_ROOT / "weights.onnx",
                    "pytorch_version": "1.15",
                    "architecture": {
                        "callable": "Model",
                        "source": "https://example.com/file.py",
                        "sha256": "0" * 64,  # dummy sha256
                    },
                },
            }
        ),
    ],
)
def test_model(model_data: Dict[str, Any], update: Dict[str, Any]):
    model_data.update(update)
    summary = validate_format(
        model_data, context=ValidationContext(perform_io_checks=False)
    )
    assert summary.status == "passed", summary.format()


def test_warn_long_name(model_data: Dict[str, Any]):
    model_data["name"] = (
        "veeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeery loooooooooooooooong name"
    )
    summary = validate_format(
        model_data, context=ValidationContext(perform_io_checks=False)
    )

    assert summary.status == "passed", summary.format()
    assert summary.details[1].warnings[0].loc == ("name",), summary.format()
    assert summary.details[1].warnings[0].msg == "Name longer than 64 characters."


def test_model_schema_raises_invalid_input_id(model_data: Dict[str, Any]):
    model_data["inputs"][0]["id"] = "invalid/id"
    summary = validate_format(
        model_data, context=ValidationContext(perform_io_checks=False)
    )
    assert summary.status == "failed", summary.format()


def test_output_fixed_shape_too_small(model_data: Dict[str, Any]):
    model_data["outputs"][0]["halo"] = 999
    summary = validate_format(
        model_data, context=ValidationContext(perform_io_checks=False)
    )
    assert summary.status == "failed", summary.format()


def test_get_axis_sizes_raises_with_surplus_n(model_data: Dict[str, Any]):
    with ValidationContext(perform_io_checks=False):
        model = ModelDescr(**model_data)

    output_tensor_id = model.inputs[0].id
    output_axis_id = AxisId("y")

    with pytest.raises(ValueError):
        _ = model.get_axis_sizes(
            ns={(output_tensor_id, output_axis_id): 1}, batch_size=1
        )


def test_get_axis_sizes_raises_with_missing_n(model_data: Dict[str, Any]):
    model_data["inputs"][0]["axes"][2] = {
        "type": "space",
        "id": "x",
        "size": {"min": 10, "step": 5},
    }

    with ValidationContext(perform_io_checks=False):
        model = ModelDescr(**model_data)

    with pytest.raises(ValueError):
        _ = model.get_axis_sizes(ns={}, batch_size=1)


def test_output_ref_shape_mismatch(model_data: Dict[str, Any]):
    model_data["outputs"][0]["axes"][2] = {
        "type": "space",
        "id": "x",
        "size": {"tensor_id": "input_1", "axis_id": "x"},
        "halo": 2,
    }
    summary = validate_format(
        model_data, context=ValidationContext(perform_io_checks=False)
    )
    assert summary.status == "passed", summary.format()
    # input_1.x -> input_1.z
    model_data["outputs"][0]["axes"][2] = {
        "type": "space",
        "id": "x",
        "size": {"tensor_id": "input_1", "axis_id": "z"},
        "halo": 2,
    }
    summary = validate_format(
        model_data, context=ValidationContext(perform_io_checks=False)
    )
    assert summary.status == "failed", summary.format()


def test_output_ref_shape_too_small(model_data: Dict[str, Any]):
    model_data["outputs"][0]["axes"][2] = {
        "type": "space",
        "id": "x",
        "size": {"tensor_id": "input_1", "axis_id": "x"},
        "halo": 2,
    }
    summary = validate_format(
        model_data, context=ValidationContext(perform_io_checks=False)
    )
    assert summary.status == "passed", summary.format()

    model_data["outputs"][0]["axes"][2]["halo"] = 999
    summary = validate_format(
        model_data, context=ValidationContext(perform_io_checks=False)
    )
    assert summary.status == "failed", summary.format()


def test_model_has_parent_with_id(model_data: Dict[str, Any]):
    model_data["parent"] = dict(id="10.5281/zenodo.5764892/6647674")
    summary = validate_format(
        model_data, context=ValidationContext(perform_io_checks=False)
    )
    assert summary.status == "passed", summary.format()


def test_model_with_expanded_output(model_data: Dict[str, Any]):
    model_data["outputs"][0]["axes"] = [
        {"type": "space", "id": "x", "size": {"tensor_id": "input_1", "axis_id": "x"}},
        {"type": "space", "id": "y", "size": {"tensor_id": "input_1", "axis_id": "y"}},
        {"type": "space", "id": "z", "size": 7},
        {"type": "channel", "channel_names": list("abc")},
    ]

    summary = validate_format(
        model_data, context=ValidationContext(perform_io_checks=False)
    )
    assert summary.status == "passed", summary.format()


def test_model_rdf_is_valid_general_rdf(model_data: Dict[str, Any]):
    model_data["type"] = "model_as_generic"
    model_data["format_version"] = "0.3.0"
    summary = validate_format(
        model_data, context=ValidationContext(perform_io_checks=False)
    )
    assert summary.status == "passed", summary.format()


def test_model_does_not_accept_unknown_fields(model_data: Dict[str, Any]):
    model_data["unknown_additional_field"] = "shouldn't be here"
    summary = validate_format(
        model_data, context=ValidationContext(perform_io_checks=False)
    )
    assert summary.status == "failed", summary.format()


def test_empty_axis_data():
    from bioimageio.spec.model.v0_5 import InputAxis as InputAxisUnion
    from bioimageio.spec.model.v0_5 import OutputAxis as OutputAxisUnion

    InputAxis = RootModel[InputAxisUnion]

    with pytest.raises(ValidationError):
        _ = InputAxis.model_validate({})

    OutputAxis = RootModel[OutputAxisUnion]

    with pytest.raises(ValidationError):
        _ = OutputAxis.model_validate({})



from pathlib import Path
from typing import Any, ClassVar, Type

from annotated_types import Predicate
from pydantic import DirectoryPath, FilePath, RootModel, StringConstraints
from typing_extensions import Annotated

from .validated_string import ValidatedString

FileName = str
AbsoluteDirectory = Annotated[DirectoryPath, Predicate(Path.is_absolute)]
AbsoluteFilePath = Annotated[FilePath, Predicate(Path.is_absolute)]

BIOIMAGEIO_YAML = "bioimageio.yaml"
ALTERNATIVE_BIOIMAGEIO_YAML_NAMES = ("rdf.yaml", "model.yaml")
ALL_BIOIMAGEIO_YAML_NAMES = (BIOIMAGEIO_YAML,) + ALTERNATIVE_BIOIMAGEIO_YAML_NAMES


class Sha256(ValidatedString):
    """SHA-256 hash value"""

    root_model: ClassVar[Type[RootModel[Any]]] = RootModel[
        Annotated[
            str,
            StringConstraints(
                strip_whitespace=True, to_lower=True, min_length=64, max_length=64
            ),
        ]
    ]



from __future__ import annotations

from contextvars import ContextVar, Token
from dataclasses import dataclass, field
from pathlib import Path
from typing import Dict, List, Optional, Union
from urllib.parse import urlsplit, urlunsplit

from pydantic import DirectoryPath

from ._settings import settings
from .io_basics import AbsoluteDirectory, FileName, Sha256
from .root_url import RootHttpUrl
from .warning_levels import WarningLevel


@dataclass(frozen=True)
class ValidationContext:
    _context_tokens: "List[Token[ValidationContext]]" = field(
        init=False, default_factory=list
    )

    root: Union[RootHttpUrl, AbsoluteDirectory] = Path()
    """url/directory serving as base to resolve any relative file paths"""

    warning_level: WarningLevel = 50
    """raise warnings of severity `s` as validation errors if `s >= warning_level`"""

    log_warnings: bool = settings.log_warnings
    """if `True` log warnings that are not raised to the console"""

    file_name: Optional[FileName] = None
    """file name of the bioimageio Yaml file"""

    perform_io_checks: bool = settings.perform_io_checks
    """wether or not to perform validation that requires file io,
    e.g. downloading a remote files.

    Existence of local absolute file paths is still being checked."""

    known_files: Dict[str, Sha256] = field(default_factory=dict)
    """allows to bypass download and hashing of referenced files"""

    def replace(
        self,
        root: Optional[Union[RootHttpUrl, DirectoryPath]] = None,
        warning_level: Optional[WarningLevel] = None,
        log_warnings: Optional[bool] = None,
        file_name: Optional[str] = None,
        perform_io_checks: Optional[bool] = None,
        known_files: Optional[Dict[str, Sha256]] = None,
    ) -> "ValidationContext":
        return ValidationContext(
            root=self.root if root is None else root,
            warning_level=(
                self.warning_level if warning_level is None else warning_level
            ),
            log_warnings=self.log_warnings if log_warnings is None else log_warnings,
            file_name=self.file_name if file_name is None else file_name,
            perform_io_checks=(
                self.perform_io_checks
                if perform_io_checks is None
                else perform_io_checks
            ),
            known_files=self.known_files if known_files is None else known_files,
        )

    def __enter__(self):
        self._context_tokens.append(validation_context_var.set(self))
        return self

    def __exit__(self, type, value, traceback):  # type: ignore
        validation_context_var.reset(self._context_tokens.pop(-1))

    @property
    def source_name(self) -> str:
        if self.file_name is None:
            return "in-memory"
        else:
            try:
                if isinstance(self.root, Path):
                    source = (self.root / self.file_name).absolute()
                else:
                    parsed = urlsplit(str(self.root))
                    path = list(parsed.path.strip("/").split("/")) + [self.file_name]
                    source = urlunsplit(
                        (
                            parsed.scheme,
                            parsed.netloc,
                            "/".join(path),
                            parsed.query,
                            parsed.fragment,
                        )
                    )
            except ValueError:
                return self.file_name
            else:
                return str(source)


validation_context_var: ContextVar[ValidationContext] = ContextVar(
    "validation_context_var", default=ValidationContext()
)



import json
from pathlib import Path

from setuptools import find_namespace_packages, setup

# Get the long description from the README file
ROOT_DIR = Path(__file__).parent.resolve()
long_description = (ROOT_DIR / "README.md").read_text(encoding="utf-8")
VERSION_FILE = ROOT_DIR / "bioimageio" / "spec" / "VERSION"
VERSION = json.loads(VERSION_FILE.read_text(encoding="utf-8"))["version"]


_ = setup(
    name="bioimageio.spec",
    version=VERSION,
    description="Parser and validator library for bioimage.io specifications",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/bioimage-io/spec-bioimage-io",
    author="bioimage.io Team",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
    packages=find_namespace_packages(exclude=["tests"]),  # Required
    install_requires=[
        "annotated-types>=0.5.0,<1",
        "email_validator",
        "imageio",
        "loguru",
        "numpy>=1.21",
        "packaging>=17.0",
        "pooch>=1.5,<2",
        "pydantic-settings",
        "pydantic>=2.7.0,<3",
        "python-dateutil",
        "python-dotenv",
        "requests",
        "rich",
        "ruyaml",
        "tqdm",
        "typing-extensions",
    ],
    extras_require={
        "dev": [
            "black",
            "deepdiff",
            "filelock",  # for session fixtures due to pytest-xdist
            "jsonschema",
            "jupyter",
            "lxml",
            "pdoc",
            "pre-commit",
            "psutil",  # parallel pytest with '-n auto'
            "pyright",
            "pytest-xdist",  # parallel pytest
            "pytest",
            "python-devtools",
            "ruff",  # check line length in cases black cannot fix it
        ]
    },
    scripts=[],
    include_package_data=True,
    project_urls={
        "Bug Reports": "https://github.com/bioimage-io/spec-bioimage-io/issues",
        "Source": "https://github.com/bioimage-io/spec-bioimage-io",
    },
)
